import io
import re
from datetime import datetime
from io import BytesIO
from typing import Generator, Optional

import fabric
from django.conf import settings

from ..models import ImportCredential
from .abstract import AbstractDataBackend
from .utils import register


@register("SFTP", save_data_in_import_source=True, passive_only=True)
class DataBackend(AbstractDataBackend):
    def __init__(
        self,
        import_credential: Optional[ImportCredential] = None,
        **kwargs,
    ):
        if not import_credential:
            raise ValueError("The sftp backend needs an import credential")

        self.username = import_credential.username
        self.password = import_credential.password
        self.host = import_credential.additional_resources.get("host", "")
        self.port = import_credential.additional_resources.get("port", "")

    def get_files(
        self,
        execution_time: datetime,
        sftp_folder,
        file_name_regex: Optional[str] = None,
        cleanup_files: Optional[bool] = getattr(settings, "WBIMPORT_EXPORT_SFTPBACKEND_CLEAN_FILES", False),
        **kwargs,
    ) -> Generator[tuple[str, BytesIO], None, None]:
        # Create Connection to the SFTP Server
        with fabric.Connection(
            host=self.host, port=self.port, user=self.username, connect_kwargs={"password": self.password}
        ) as conn:
            sftp = conn.sftp()

            # Change working directory
            sftp.chdir(sftp_folder)

            # Filter all the files we need
            file_names = sftp.listdir()
            if file_name_regex:
                # Compile the regex for later filtering the files
                file_names = filter(lambda x: re.match(file_name_regex, x), file_names)

            for file_name in file_names:
                # Create a Buffer where we write the file to
                sftp_file = io.BytesIO()
                sftp.getfo(file_name, sftp_file)
                yield file_name, sftp_file

                if cleanup_files:
                    # Delete the file from the server
                    sftp.remove(file_name)



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from . import viewsets

router = WBCoreRouter()

# Representations
router.register(r"entryrepresentation", viewsets.EntryRepresentationViewSet, basename="entryrepresentation")
router.register(r"personrepresentation", viewsets.PersonRepresentationViewSet, basename="personrepresentation")
router.register(
    r"personinchargerepresentation",
    viewsets.PersonInChargeRepresentationViewSet,
    basename="personinchargerepresentation",
)
router.register(
    r"emailcontactrepresentation", viewsets.EmailContactRepresentationViewSet, basename="emailcontactrepresentation"
)
router.register(
    r"addresscontactrepresentation", viewsets.AddressContactViewSet, basename="addresscontactrepresentation"
)
router.register(
    r"socialmediacontactrepresentation",
    viewsets.SocialMediaContactRepresentationViewSet,
    basename="socialmediacontactrepresentation",
)
router.register(
    r"telephonecontactrepresentation",
    viewsets.TelephoneContactRepresentationViewSet,
    basename="telephonecontactrepresentation",
)
router.register(
    r"bankingcontactrepresentation",
    viewsets.BankingContactRepresentationViewSet,
    basename="bankingcontactrepresentation",
)
router.register(
    r"websitecontactrepresentation",
    viewsets.WebsiteContactRepresentationViewSet,
    basename="websitecontactrepresentation",
)

# Relationships
router.register(r"relationship", viewsets.RelationshipModelViewSet, basename="relationship")
router.register(r"relationship-type", viewsets.RelationshipTypeModelViewSet, basename="relationship-type")
router.register(
    r"relationshiprepresentation", viewsets.RelationshipRepresentationViewSet, basename="relationshiprepresentation"
)
router.register(
    r"relationshiptyperepresentation",
    viewsets.RelationshipTypeRepresentationViewSet,
    basename="relationshiptyperepresentation",
)
router.register(
    r"employeremployeerelationship",
    viewsets.EmployerEmployeeRelationshipModelViewSet,
    basename="employeremployeerelationship",
)
router.register(r"clientmanagerrelationship", viewsets.ClientManagerViewSet, basename="clientmanagerrelationship")
router.register(
    r"clientmanagerrelationshiprepresentation",
    viewsets.ClientManagerRelationshipRepresentationViewSet,
    basename="clientmanagerrelationshiprepresentation",
)
router.register(
    r"clientmanagerrelationship-usermanager",
    viewsets.UserIsManagerViewSet,
    basename="clientmanagerrelationship-usermanager",
)
router.register(
    r"clientmanagerrelationship-userclient",
    viewsets.UserIsClientViewSet,
    basename="clientmanagerrelationship-userclient",
)

# Entry
router.register(r"entry", viewsets.EntryModelViewSet, basename="entry")
router.register(r"person", viewsets.PersonModelViewSet, basename="person")
router.register(r"bank", viewsets.BankModelViewSet, basename="bank")
router.register(r"systememployee", viewsets.SystemEmployeeModelViewSet, basename="systememployee")
router.register(r"company", viewsets.CompanyModelViewSet, basename="company")
router.register(r"companyrepresentation", viewsets.CompanyRepresentationViewSet, basename="companyrepresentation")
# Contacts
router.register(r"addresscontact", viewsets.AddressContactViewSet, basename="addresscontact")
router.register(r"bankingcontact", viewsets.BankingContactViewSet, basename="bankingcontact")
router.register(r"emailcontact", viewsets.EmailContactViewSet, basename="emailcontact")
router.register(r"socialmediacontact", viewsets.SocialMediaContactViewSet, basename="socialmediacontact")
router.register(r"telephonecontact", viewsets.TelephoneContactViewSet, basename="telephonecontact")
router.register(r"websitecontact", viewsets.WebsiteContactViewSet, basename="websitecontact")


# Utils
router.register(r"customerstatus", viewsets.CustomerStatusModelViewSet, basename="customerstatus")
router.register(
    r"customerstatusrepresentation",
    viewsets.CustomerStatusRepresentationViewSet,
    basename="customerstatusrepresentation",
)
router.register(r"position", viewsets.PositionModelViewSet, basename="position")
router.register(r"positionrepresentation", viewsets.PositionRepresentationViewSet, basename="positionrepresentation")
router.register(r"companytype", viewsets.CompanyTypeModelViewSet, basename="companytype")
router.register(
    r"companytyperepresentation", viewsets.CompanyTypeRepresentationViewSet, basename="companytyperepresentation"
)
router.register(r"specialization", viewsets.SpecializationModelViewSet, basename="specialization")
router.register(
    r"specializationrepresentation",
    viewsets.SpecializationRepresentationViewSet,
    basename="specializationrepresentation",
)

employer_router = WBCoreRouter()
employer_router.register(r"employee", viewsets.EmployeeEmployerModelViewSet, basename="employer-employee")

employee_router = WBCoreRouter()
employee_router.register(r"employer", viewsets.EmployerEmployeeModelViewSet, basename="employee-employer")

entry_router = WBCoreRouter()
entry_router.register(r"emailcontact", viewsets.EmailContactEntryViewSet, basename="entry-emailcontact")
entry_router.register(r"addresscontact", viewsets.AddressContactEntryViewSet, basename="entry-addresscontact")
entry_router.register(r"telephonecontact", viewsets.TelephoneContactEntryViewSet, basename="entry-telephonecontact")
entry_router.register(r"websitecontact", viewsets.WebsiteContactEntryViewSet, basename="entry-websitecontact")
entry_router.register(r"bankingcontact", viewsets.BankingContactEntryViewSet, basename="entry-bankingcontact")
entry_router.register(
    r"socialmediacontact", viewsets.SocialMediaContactEntryViewSet, basename="entry-socialmediacontact"
)
entry_router.register(r"relationship", viewsets.RelationshipModelEntryViewSet, basename="entry-relationship")

urlpatterns = [
    path("", include(router.urls)),
    path("company/<int:employer_id>/", include(employer_router.urls)),
    path("person/<int:employee_id>/", include(employee_router.urls)),
    path("entry/<int:entry_id>/", include(entry_router.urls)),
]



from wbcore.configs.registry import ConfigRegistry
from wbcore.contrib.gleap.hashes import get_hash_from_user


def test_different_user_id():
    hash1 = get_hash_from_user(1, "abc")
    hash2 = get_hash_from_user(2, "abc")

    assert hash1 != hash2


def test_same_user_id():
    hash1 = get_hash_from_user(1, "abc")
    hash2 = get_hash_from_user(1, "abc")

    assert hash1 == hash2


def test_different_secret():
    hash1 = get_hash_from_user(1, "abc")
    hash2 = get_hash_from_user(1, "abc1")

    assert hash1 != hash2


def test_gleap_config(config_registry: ConfigRegistry):
    gleap = config_registry.get_config_dict()["gleap"]
    assert gleap["user_identity_endpoint"]
    assert gleap["api_token"]



from rest_framework.renderers import BrowsableAPIRenderer


class BrowsableAPIRendererWithoutForms(BrowsableAPIRenderer):
    """Renders the browsable api, but excludes the forms."""

    def get_rendered_html_form(self, data, view, method, request):
        """Why render _any_ forms at all. This method should return
        rendered HTML, so let's simply return an empty string.
        """
        if method in ("DELETE", "OPTIONS"):
            return True  # Don't actually need to return a form
        return ""



from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig


class PersonModelTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Persons")

    def get_instance_title(self):
        return _("Person: {{first_name}} {{last_name}}")

    def get_create_title(self):
        return _("New Person")


class CompanyModelTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Company: {{name}}")

    def get_list_title(self):
        return _("Companies")

    def get_create_title(self):
        return _("New Company")


class UserIsManagerTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Your Clients/Prospects/Contacts")



import factory
from wbcore.contrib.example_app.models import Role


class RoleFactory(factory.django.DjangoModelFactory):
    title = factory.Sequence(lambda n: "Role %d" % n)

    class Meta:
        model = Role



from wbcore.metadata.configs.endpoints import EndpointViewConfig


class StepEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs) -> str:
        if self.view.kwargs.get("workflow_id"):
            return ""
        return super().get_create_endpoint(**kwargs)

    def _get_instance_endpoint(self, **kwargs) -> str:
        return "{{casted_endpoint}}"

    def get_update_endpoint(self, **kwargs):
        if self.view.kwargs.get("workflow_id"):
            return ""
        return super().get_update_endpoint(**kwargs)


class DisplayEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs) -> None:
        return None

    def get_instance_endpoint(self, **kwargs) -> None:
        return None



from django.shortcuts import get_object_or_404
from rest_framework import permissions
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.workflow.filters import WorkflowFilter
from wbcore.contrib.workflow.models import StartStep, Workflow
from wbcore.contrib.workflow.serializers import WorkflowModelSerializer
from wbcore.contrib.workflow.viewsets.display import WorkflowDisplayConfig
from wbcore.contrib.workflow.viewsets.titles import WorkflowTitleConfig


class WorkflowModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/workflow.md"
    queryset = Workflow.objects.all()
    serializer_class = WorkflowModelSerializer
    display_config_class = WorkflowDisplayConfig
    title_config_class = WorkflowTitleConfig
    search_fields = ordering = ("name",)
    filterset_class = WorkflowFilter
    ordering_fields = (
        "name",
        "single_instance_execution",
        "model__model",
        "status_field",
        "preserve_instance",
    )

    @action(detail=True, methods=["PATCH"], permission_classes=[permissions.IsAuthenticated])
    def start(self, request, pk=None):
        start_step = get_object_or_404(StartStep, pk=request.GET.get("step_id"))
        workflow = get_object_or_404(Workflow, pk=pk)
        instance_id: int | None = request.GET.get("instance_id")
        instance = get_object_or_404(workflow.model.model_class(), pk=instance_id) if instance_id else None
        workflow.start_workflow(start_step, instance)
        return Response({"status": "ok"})



from typing import Iterator

import pytest
from wbcore.contrib.dataloader.tests.test.dataloaders.proxies import (
    EntityTestDataloaderProxy,
)
from wbcore.contrib.dataloader.tests.test.models import EntityTest


@pytest.mark.django_db
class TestEntityTestDataloader:
    def test_dataloader_proxy(self, entity_test):
        dl_proxy = EntityTest.objects.filter(id=entity_test.id).dl
        assert isinstance(dl_proxy, EntityTestDataloaderProxy), "The proxy should exist"

    def test_dataloader(self, entity_test):
        result = EntityTest.objects.filter(id=entity_test.id).dl.data(10)
        assert isinstance(result, Iterator), "The result of a dataloader should be an iterator"
        assert len(list(result)) == 10, "The result should have a length of 10 items"
        assert all(isinstance(item, int) for item in result), "All items in the result list should be integers"

    def test_multiple_entries_dataloader(self, entity_test_factory):
        entity_test_factory.create()
        entity_test_factory.create()
        entity_test_factory.create()
        result = EntityTest.objects.all().dl.data(10)
        assert len(list(result)) == 30, "The result should have a length of 30 (3 EntityTest x 10) items"

    def test_multiple_dataloader(self, entity_test, entity_test_over_1000):
        result = EntityTest.objects.all().dl.data(10)
        assert (
            len(list(result)) == 20
        ), "The result should have a length of 20 (1 EntityTest x 2 Dataloader x 10) items"

    def test_empty_dataloader(self):
        result = list(EntityTest.objects.all().dl.data(10))
        assert len(result) == 0, "Having no entities, should result in an empty list"



from wbcore import serializers
from wbcore.content_type.serializers import (
    ContentTypeRepresentationSerializer,
    DynamicObjectIDRepresentationSerializer,
)
from wbcore.contrib.documents.models import DocumentModelRelationship


class DocumentModelRelationshipModelSerializer(serializers.ModelSerializer):
    _content_type = ContentTypeRepresentationSerializer(
        source="content_type",
        label_key="{{model_title}}",
    )
    _object_id = DynamicObjectIDRepresentationSerializer(
        source="object_id",
        optional_get_parameters={"content_type": "content_type"},
        depends_on=[{"field": "content_type", "options": {}}],
    )

    class Meta:
        model = DocumentModelRelationship
        fields = (
            "id",
            "content_type",
            "_content_type",
            "object_id",
            "_object_id",
            "content_object_repr",
            "document",
            "primary",
        )



import operator
from functools import reduce

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Q
from wbcore import serializers


class ContentTypeRepresentationSerializer(serializers.RepresentationSerializer):
    endpoint = "wbcore:contenttyperepresentation-list"
    value_key = "id"

    model_title = serializers.SerializerMethodField()

    def __init__(self, *args, model_callback_title=None, label_key="{{app_label}} | {{model}}", **kwargs):
        self.allowed_types = kwargs.pop("allowed_types", ContentType.objects.none())

        # for ease of use, we allow the allowed_type attribute to be set as a list of app_label and model (as tuple). In that case, we extract the valid content type queryset base on the join of conditions
        if isinstance(self.allowed_types, list):
            _conditions = [
                (Q(app_label=allowed_type[0]) & Q(model__iexact=allowed_type[1]))
                for allowed_type in self.allowed_types
                if isinstance(allowed_type, tuple) and len(allowed_type) == 2
            ]
            if _conditions:
                self.allowed_types = ContentType.objects.filter(reduce(operator.or_, _conditions))
            else:
                self.allowed_types = ContentType.objects.none()
        self.model_callback_title = model_callback_title
        super().__init__(*args, label_key=label_key, **kwargs)

    def get_model_title(self, obj):
        return obj.model.title()

    def _get_filter_params(self, request):
        filter_params = super()._get_filter_params(request)
        if self.allowed_types.exists():
            filter_params["id__in"] = ",".join(map(str, self.allowed_types.values_list("id", flat=True)))
        return filter_params

    class Meta:
        model = ContentType
        read_only_fields = (
            "app_label",
            "model",
            "model_title",
        )
        fields = ("id", "app_label", "model", "model_title")


class GenericModel(models.Model):
    id = models.IntegerField(primary_key=True)
    label = models.CharField(max_length=256)

    class Meta:
        managed = False


class DynamicObjectIDRepresentationSerializer(serializers.RepresentationSerializer):
    endpoint = "wbcore:dynamiccontenttyperepresentation-list"
    value_key = "{{id}}"
    label_key = "{{label}}"

    label = serializers.CharField(read_only=True)
    id = serializers.CharField(read_only=True)

    def __init__(self, *args, content_type_field_name: str = "content_type", **kwargs):
        self.content_type_field_name = content_type_field_name
        super().__init__(*args, **kwargs)

    def to_representation(self, value):
        representation = super().to_representation(value)
        representation["id"] = getattr(value, "pk", value)
        try:
            if content_type := getattr(self.parent.instance, self.content_type_field_name, None):
                obj = content_type.get_object_for_this_type(id=value)
                representation["label"] = str(obj)
        except Exception:
            representation["label"] = str(representation["id"])
        return representation

    class Meta:
        model = GenericModel
        read_only_fields = (
            "id",
            "label",
        )
        fields = ("id", "label")



from configurations import values


class CeleryTest:
    CELERY_TASK_ALWAYS_EAGER = True


class Celery:
    CELERY_BROKER_TRANSPORT_OPTIONS = {
        "max_retries": 3,
        "interval_start": 0,
        "interval_step": 0.2,
        "interval_max": 0.5,
    }
    CELERY_BROKER_URL = values.Value("redis://localhost:6379", environ_name="REDIS_URL", environ_prefix=None)
    CELERY_BROKER_CONNECTION_RETRY_ON_STARTUP = values.BooleanValue(True, environ_prefix=None)
    CELERY_TASK_SERIALIZER = values.Value("pickle", environ_prefix=None)
    CELERY_ACCEPT_CONTENT = values.ListValue(["json", "pickle"], environ_prefix=None)
    CELERY_RESULT_SERIALIZER = values.Value("json", environ_prefix=None)
    CELERY_BEAT_SCHEDULER = "django_celery_beat.schedulers:DatabaseScheduler"
    CELERY_RESULT_BACKEND = values.Value("redis://localhost:6379", environ_name="REDIS_URL", environ_prefix=None)

    if "rediss://" in CELERY_BROKER_URL:
        CELERY_BROKER_USE_SSL = {"ssl_cert_reqs": "none"}
    else:
        CELERY_BROKER_USE_SSL = {}



# Generated by Django 5.0.3 on 2024-04-24 11:44

import django.core.serializers.json
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


def migrate_status(apps, schema_editor):
    ImportSource = apps.get_model("io", "ImportSource")
    ImportSource.objects.filter(status="IMPORTED").update(status="PROCESSED")


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("io", "0003_auto_20240103_1000"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(migrate_status),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.AlterField(
            model_name="importsource",
            name="status",
            field=models.CharField(
                choices=[
                    ("PENDING", "Pending"),
                    ("PROCESSED", "Processed"),
                    ("WARNING", "Warning"),
                    ("ERROR", "Error"),
                    ("IGNORED", "Ignore"),
                ],
                default="PENDING",
                max_length=16,
            ),
        ),
        migrations.AddField(
            model_name="importsource",
            name="errors_log",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.CreateModel(
            name="ExportSource",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("last_updated", models.DateTimeField(auto_now=True)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("PENDING", "Pending"),
                            ("PROCESSED", "Processed"),
                            ("WARNING", "Warning"),
                            ("ERROR", "Error"),
                            ("IGNORED", "Ignore"),
                        ],
                        default="PENDING",
                        max_length=16,
                    ),
                ),
                ("log", models.TextField(blank=True, null=True)),
                ("origin", models.CharField(blank=True, max_length=255, null=True)),
                ("file", models.FileField(blank=True, max_length=256, null=True, upload_to="io/export_source/files")),
                (
                    "format",
                    models.IntegerField(
                        choices=[
                            (0, "CSV"),
                            (1, "XLS"),
                            (2, "XLSX"),
                            (3, "TSV"),
                            (4, "ODS"),
                            (5, "JSON"),
                            (6, "YAML"),
                            (7, "HTML"),
                        ],
                        default=0,
                        verbose_name="Format of file to be exported",
                    ),
                ),
                (
                    "resource_path",
                    models.CharField(default="", max_length=255, verbose_name="Resource path to use when exporting"),
                ),
                ("resource_kwargs", models.JSONField(blank=True, default=dict)),
                ("query_str", models.TextField(verbose_name="SQL query to be executed")),
                (
                    "query_params",
                    models.JSONField(
                        blank=True,
                        default=list,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        verbose_name="SQL query parameters to be used with the sql query",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                        verbose_name="Export job Content Type",
                    ),
                ),
                (
                    "creator",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Creator",
                    ),
                ),
            ],
            options={
                "verbose_name": "Export Source",
                "verbose_name_plural": "Export Sources",
                "notification_types": [
                    (
                        "io.export_done",
                        "File Export has finished",
                        "Notifies user when their submitted export files is done and available",
                        True,
                        True,
                        True,
                    )
                ],
            },
        ),
    ]



import random

import pytest
from faker import Faker
from wbcore import serializers as wb_serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import (
    ActionButton,
    DropDownButton,
    HyperlinkButton,
    WidgetButton,
)
from wbcore.metadata.configs.buttons.enums import ButtonDefaultColor, HyperlinkTarget
from wbcore.metadata.configs.display.instance_display import create_simple_display
from wbcore.metadata.utils import prefix_key

fake = Faker()


class TestDropDownButton:
    @pytest.fixture()
    def button(self):
        return DropDownButton(
            icon=random.choice(list(WBIcon)).icon,
            title=fake.sentence(),
            color=random.choice(list(ButtonDefaultColor)).value,
            buttons=[WidgetButton(key=fake.word(), label=fake.sentence())],
        )

    def test_serialize(self, rf, button):
        serialized_btn = button.serialize(rf)
        assert serialized_btn["icon"] == button.icon
        assert serialized_btn["title"] == button.title
        assert serialized_btn["color"] == button.color
        assert serialized_btn["level"] == button.color
        assert serialized_btn["buttons"] == [button.buttons[0].serialize(rf)]
        assert serialized_btn["type"] == DropDownButton.button_type.value

    @pytest.mark.parametrize("key_prefix", [fake.word()])
    def test_serialize_with_prefix(self, rf, button, key_prefix):
        serialized_btn = button.serialize(rf, key_prefix=key_prefix)
        nested_button = button.buttons[0]

        assert serialized_btn["icon"] == button.icon
        assert serialized_btn["title"] == button.title
        assert serialized_btn["color"] == button.color
        assert serialized_btn["level"] == button.color
        assert serialized_btn["buttons"] == [nested_button.serialize(rf, key_prefix=key_prefix)]
        assert serialized_btn["type"] == DropDownButton.button_type.value
        assert serialized_btn["buttons"][0]["key"] == prefix_key(nested_button.key, key_prefix)

    def test_hash(self, button):
        assert hash(button) == hash(button.title)


class TestWidgetButton:
    button_class = WidgetButton

    @pytest.fixture()
    def key_button(self):
        return self.button_class(
            icon=random.choice(list(WBIcon)).icon,
            title=fake.sentence(),
            color=random.choice(list(ButtonDefaultColor)).value,
            key=fake.word(),
        )

    @pytest.fixture()
    def endpoint_button(self):
        return self.button_class(
            icon=random.choice(list(WBIcon)).icon,
            title=fake.sentence(),
            color=random.choice(list(ButtonDefaultColor)).value,
            endpoint=fake.url(),
        )

    def test_serialize(self, rf, key_button):
        serialized_btn = key_button.serialize(rf)
        assert serialized_btn["icon"] == key_button.icon
        assert serialized_btn["title"] == key_button.title
        assert serialized_btn["color"] == key_button.color
        assert serialized_btn["level"] == key_button.color
        assert serialized_btn["key"] == key_button.key
        assert serialized_btn["type"] == self.button_class.button_type.value

    @pytest.mark.parametrize("key_prefix", [fake.word()])
    def test_serialize_with_prefix(self, rf, key_button, key_prefix):
        serialized_btn = key_button.serialize(rf, key_prefix=key_prefix)

        assert serialized_btn["icon"] == key_button.icon
        assert serialized_btn["title"] == key_button.title
        assert serialized_btn["color"] == key_button.color
        assert serialized_btn["level"] == key_button.color
        assert serialized_btn["type"] == self.button_class.button_type.value
        assert serialized_btn["key"] == prefix_key(key_button.key, key_prefix)

    def test_key_hash(self, key_button):
        assert hash(key_button) == hash(key_button.key)

    def test_endpoint_hash(self, endpoint_button):
        assert hash(endpoint_button) == hash(endpoint_button.endpoint)

    def test_serialize_endpoint_button(self, rf, endpoint_button):
        serialized_btn = endpoint_button.serialize(rf)
        assert serialized_btn["icon"] == endpoint_button.icon
        assert serialized_btn["title"] == endpoint_button.title
        assert serialized_btn["color"] == endpoint_button.color
        assert serialized_btn["level"] == endpoint_button.color
        assert serialized_btn["endpoint"] == endpoint_button.endpoint
        assert serialized_btn["type"] == self.button_class.button_type.value

        base_endpoint = endpoint_button.endpoint
        endpoint_button.new_mode = True
        serialized_btn = endpoint_button.serialize(rf)
        assert serialized_btn["endpoint"] == base_endpoint + "?new_mode=true"
        assert serialized_btn["new_mode"] is True


class TestHyperlinkButtonutton(TestWidgetButton):
    button_class = HyperlinkButton

    def test_serialize(self, rf, key_button):
        super().test_serialize(rf, key_button)
        key_button.target = random.choice(list(HyperlinkTarget))
        serialized_btn = key_button.serialize(rf)
        assert serialized_btn["target"] == key_button.target.value

    def test_serialize_endpoint_button(self, rf, endpoint_button):
        serialized_btn = endpoint_button.serialize(rf)
        assert serialized_btn["icon"] == endpoint_button.icon
        assert serialized_btn["title"] == endpoint_button.title
        assert serialized_btn["color"] == endpoint_button.color
        assert serialized_btn["level"] == endpoint_button.color
        assert serialized_btn["endpoint"] == endpoint_button.endpoint
        assert serialized_btn["type"] == self.button_class.button_type.value

        # No new mode supported for hyperlink


class TestActionButtonButton(TestHyperlinkButtonutton):
    button_class = ActionButton

    @pytest.fixture()
    def key_button(self):
        class BaseSerializer(wb_serializers.Serializer):
            field = wb_serializers.CharField()

            class Meta:
                fields = ("field",)

        return ActionButton(
            icon=random.choice(list(WBIcon)).icon,
            title=fake.sentence(),
            color=random.choice(list(ButtonDefaultColor)).value,
            key=fake.word(),
            method=random.choice(list(RequestType)),
            action_label=fake.word(),
            description_fields=fake.sentence(),
            instance_display=create_simple_display([["field"]]),
            serializer=BaseSerializer,
            identifiers=(fake.word(),),
        )

    def test_serialize(self, rf, key_button):
        TestWidgetButton.test_serialize(self, rf, key_button)
        serialized_btn = key_button.serialize(rf)

        assert serialized_btn["action_label"] == key_button.action_label
        assert serialized_btn["method"] == key_button.method.value
        assert serialized_btn["description_fields"] == key_button.description_fields
        assert serialized_btn["confirm_config"] == key_button.confirm_config.serialize(rf)
        assert serialized_btn["cancel_config"] == key_button.cancel_config.serialize(rf)
        assert serialized_btn["identifiers"] == key_button.identifiers
        assert serialized_btn["instance_display"] == key_button.instance_display.serialize()
        assert (
            serialized_btn["fields"]["field"]
            == key_button.serializer().fields["field"].get_representation(rf, "field")[1]
        )



from wbcore.metadata.configs.display.instance_display.layouts.layouts import Layout


class TestLayout:
    def test_minimal_init(self):
        assert Layout([[]]) is not None

    def test_minimal_serialize(self):
        assert Layout([[]]).serialize() == {
            "gridTemplateAreas": "''",
            "sections": [],
            "inlines": [],
            "gap": "10px",
            "gridAutoRows": "min-content",
        }



from django.core.cache import cache
from django.http import HttpRequest
from rest_framework import permissions, status
from rest_framework.decorators import (
    api_view,
    authentication_classes,
    permission_classes,
)
from rest_framework.response import Response
from wbcore.contrib.authentication.authentication import JWTCookieAuthentication


@api_view(["PATCH"])
@permission_classes([permissions.IsAdminUser])
@authentication_classes([JWTCookieAuthentication])
def clear_cache(request: HttpRequest, cache_key: str, *args, **kwargs):
    cache.delete(cache_key)
    return Response(status=status.HTTP_204_NO_CONTENT)



import pytest
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.authentication.models import User
from wbcore.test import find_element, login

USER_PASSWORD = "User_Password"


@pytest.mark.skip(reason="no way of currently testing this")
@pytest.mark.django_db
class TestE2ELogin:
    def test_login_success(self, live_server, selenium):
        user: User = SuperUserFactory(plaintext_password=USER_PASSWORD)
        selenium.get(live_server.url)
        assert not find_element(selenium, "//div[contains(@class, 'sidebar')]")
        login(selenium, user.email, USER_PASSWORD)
        assert find_element(selenium, "//div[contains(@class, 'sidebar')]")

    def test_login_failure(self, live_server, selenium):
        user: User = SuperUserFactory(plaintext_password=USER_PASSWORD)
        selenium.get(live_server.url)
        assert not find_element(selenium, "//p[@type='error']")
        login(selenium, user.email + "Wrong Name", USER_PASSWORD)
        assert find_element(selenium, "//p[@type='error']")



from typing import Callable


def register(serializer_class=None):
    """
    Register the given model as a "workflowable" model
    """
    from wbcore.contrib.workflow.sites import workflow_site

    def _model_wrapper(model_class):
        workflow_site.registered_model_classes_serializer_map[model_class] = serializer_class
        return model_class

    return _model_wrapper


def register_assignee(name: str) -> Callable:
    from wbcore.contrib.workflow.sites import workflow_site

    def _wrapper(func: Callable) -> Callable:
        workflow_site.registered_assignees_methods[func.__name__] = func
        workflow_site.registered_assignees_names[func.__name__] = name
        return func

    return _wrapper



from contextlib import suppress

import graphviz
from django.contrib.contenttypes.models import ContentType
from django.core.files.base import ContentFile
from django.db import models, transaction
from django.db.models import SET_NULL, Q, QuerySet
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from wbcore.contrib.authentication.models import User
from wbcore.contrib.workflow.models import (
    Condition,
    DataValue,
    DecisionStep,
    EmailStep,
    FinishStep,
    JoinStep,
    Process,
    ProcessStep,
    ScriptStep,
    SplitStep,
    StartStep,
    Step,
    Transition,
    UserStep,
    activate_step,
)
from wbcore.contrib.workflow.utils import get_model_serializer_class_for_instance
from wbcore.models import WBModel
from wbcore.serializers.serializers import ModelSerializer
from wbcore.utils.signals import DisconnectSignals, SignalWrapper


class Workflow(WBModel):
    """The base workflow model that can be attached to a generic model to update its status field."""

    name = models.CharField(max_length=128, verbose_name=_("Name"), unique=True)
    single_instance_execution = models.BooleanField(
        default=False,
        verbose_name=_("Single Instance Execution"),
        help_text=_(
            "Indicates wether the workflow is restricted to only one running workflow at a time. If an instance is attached to the workflow this will restrict the workflow execution per model instance."
        ),
    )
    model = models.ForeignKey(to=ContentType, on_delete=SET_NULL, null=True, blank=True, verbose_name=_("Model"))
    status_field = models.CharField(
        max_length=64,
        verbose_name=_("Status Field"),
        help_text=_("The status field name in the attached model that needs to be updated by the workflow"),
        null=True,
        blank=True,
    )
    preserve_instance = models.BooleanField(
        default=False,
        verbose_name=_("Preserve Instance"),
        help_text=_(
            "This will preserve the attached instance's state in a JSON field when starting the workflow. You then have the option to write this data back onto the instance in a finish step, effectively preserving the instance's state throughout the workflow."
        ),
    )
    graph = models.ImageField(verbose_name=_("Graph"), blank=True, null=True)

    def start_workflow(self, start_step: StartStep, instance=None):
        """Starts a new workflow

        Args:
            start_step (StartStep): The workflow's start step
            instance (_type_, optional): An attached instance
        """

        if instance and self.preserve_instance:
            serializer: ModelSerializer = get_model_serializer_class_for_instance(instance)
            data: dict = serializer(instance).data
            process_instance = Process.objects.create(workflow=self, instance=instance, preserved_instance=data)
        else:
            process_instance = Process.objects.create(workflow=self, instance=instance)

        for data in self.attached_data.all():
            if data.default:
                DataValue.objects.create(
                    data=data,
                    value=data.default,
                    process=process_instance,
                )
            else:
                DataValue.objects.create(data=data, process=process_instance)
        transaction.on_commit(lambda: activate_step.delay(start_step.pk, process_instance.pk))

    def get_start_steps_for_workflow(self) -> QuerySet[StartStep]:
        return StartStep.objects.filter(
            Q(workflow__model__isnull=True)
            & Q(workflow=self)
            & ~Q(Q(workflow__single_instance_execution=True) & Q(workflow__process__state=Process.ProcessState.ACTIVE))
        )

    def generate_workflow_png(self) -> bytes:
        """Generates a PNG file visually displaying the workflow

        Returns:
            bytes: The created PNG as bytes

        Raises:
            graphviz.backend.execute.ExecutableNotFound: When graphviz executable is not installed on target system
        """

        # Initialize graph
        dot = graphviz.Digraph(
            "workflow",
            graph_attr={"rankdir": "LR"},
            node_attr={
                "shape": "rectangle",
                "style": "rounded",
                "orientation": "0",
                "width": "1.0",
                "height": "0.7",
                "fontname": "times bold",
                "margin": "0.0, 0.0",
            },
            edge_attr={
                "fontsize": "10.0",
                "fontname": "times italic",
            },
        )

        # Add nodes (steps)
        for step in self.associated_steps.all():
            if step.step_type == Step.StepType.FINISHSTEP:
                dot.node(
                    str(step.code),
                    step.name,
                    shape="circle",
                    style="bold",
                    width="0.5",
                    height="0.5",
                )
            elif step.step_type == Step.StepType.STARTSTEP:
                dot.node(
                    str(step.code),
                    step.name,
                    shape="circle",
                    width="0.5",
                    height="0.5",
                )
            elif step.step_type in [Step.StepType.DECISIONSTEP, Step.StepType.JOINSTEP, Step.StepType.SPLITSTEP]:
                dot.node(
                    str(step.code),
                    step.name,
                    shape="square",
                    orientation="45",
                    style="solid",
                )
            else:
                dot.node(
                    str(step.code),
                    step.name,
                )

            # Add incoming edges (transitions) for node
            for transition in step.get_incoming_transitions():
                edge_label = ""

                # Use condition as edge label. Can sadly only properly display one condition per edge
                if condition := transition.associated_conditions.first():
                    edge_label = gettext("if {} {} {}").format(
                        condition.attribute_name, condition.operator, condition.expected_value
                    )

                dot.edge(
                    str(transition.from_step.code),
                    str(transition.to_step.code),
                    label=edge_label,
                )

        return dot.pipe(format="png")

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:workflow"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:workflowrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    @classmethod
    def get_start_steps_for_instance(cls, instance) -> list[StartStep]:
        content_type_for_instance = ContentType.objects.get_for_model(instance)

        start_steps = StartStep.objects.filter(
            Q(workflow__model=content_type_for_instance)
            & ~Q(
                Q(workflow__single_instance_execution=True)
                & Q(
                    workflow__process__in=Process.objects.filter(
                        finished__isnull=True, content_type=content_type_for_instance, instance_id=instance.id
                    )
                )
            )
        )

        valid_start_steps = []
        for step in start_steps:
            if hasattr(instance, step.workflow.status_field):
                instance_status = getattr(instance, step.workflow.status_field)
                if step.status == instance_status:
                    valid_start_steps.append(step)

        return valid_start_steps

    @classmethod
    def get_next_user_step_transitions_for_instance(
        cls, instance, user: User
    ) -> list[tuple[list[Transition], ProcessStep]]:
        # TODO: OPTIMIZE
        process_steps = ProcessStep.objects.filter(
            state=ProcessStep.StepState.ACTIVE,
            process__instance_id=instance.pk,
            process__content_type=ContentType.objects.get_for_model(instance),
            step__step_type=Step.StepType.USERSTEP,
        )
        # User has to be assigned
        if not user.is_superuser:
            process_steps = process_steps.filter(
                Q(assignee=user)
                | (
                    Q(group__user=user)
                    & Q(step__in=UserStep.objects.filter(assignee_method__isnull=True))
                    & Q(assignee__isnull=True)
                )
            )

        valid_transitions_and_process_steps = []
        for process_step in process_steps:
            transitions = Transition.objects.filter(from_step__process_steps=process_step)
            # User needs to have permission to view the step
            if not user.is_superuser:
                transitions = transitions.filter(
                    Q(
                        Q(Q(Q(from_step__permission__user=user) | Q(from_step__permission__group__user=user)))
                        | Q(from_step__permission__isnull=True)
                    )
                )
            # Check conditions
            valid_transitions = []
            for transition in transitions:
                if transition.all_conditions_satisfied(process_step):
                    valid_transitions.append(transition)

            if valid_transitions:
                valid_transitions_and_process_steps.append(
                    tuple(
                        (
                            valid_transitions,
                            process_step,
                        )
                    ),
                )

        return valid_transitions_and_process_steps

    class Meta:
        verbose_name = _("Workflow")
        verbose_name_plural = _("Workflows")


@receiver(post_save, sender=Condition)
@receiver(post_save, sender=Transition)
@receiver(post_save, sender=UserStep)
@receiver(post_save, sender=EmailStep)
@receiver(post_save, sender=DecisionStep)
@receiver(post_save, sender=SplitStep)
@receiver(post_save, sender=JoinStep)
@receiver(post_save, sender=ScriptStep)
@receiver(post_save, sender=StartStep)
@receiver(post_save, sender=FinishStep)
@receiver(post_save, sender=Workflow)
def re_render_workflow_graph(sender, instance: Workflow | Step | Transition | Condition, raw: bool, **kwargs):
    if not raw:
        match sender.__name__:
            case "Workflow":
                workflow = instance
            case "UserStep" | "EmailStep" | "DecisionStep" | "SplitStep" | "JoinStep" | "ScriptStep" | "StartStep" | "FinishStep":
                workflow = instance.workflow
            case "Transition":
                workflow = instance.to_step.workflow
            case "Condition":
                workflow = instance.transition.to_step.workflow
            case _:
                return
        with suppress(graphviz.backend.execute.ExecutableNotFound):
            graph = ContentFile(workflow.generate_workflow_png(), "workflow_graph")
            if workflow.graph:
                workflow.graph.delete(save=False)
            workflow.graph = graph
            with DisconnectSignals([SignalWrapper(post_save, re_render_workflow_graph, sender)]):
                workflow.save()



import factory
from wbcore.contrib.directory.factories import CompanyFactory
from wbcore.contrib.example_app.models import Team, TeamResults


class TeamFactory(CompanyFactory):
    coach = factory.SubFactory("wbcore.contrib.example_app.factories.SportPersonFactory")
    home_stadium = factory.SubFactory("wbcore.contrib.example_app.factories.StadiumFactory")
    founded_date = factory.Faker("date_object")
    # city = factory.SubFactory("wbcore.contrib.geography.factories.CityFactory")

    @factory.post_generation
    def opponents(self, create, extracted, **kwargs):
        if extracted:
            for team in extracted:
                self.opponents.add(team)

    class Meta:
        model = Team


class TeamResultsFactory(factory.django.DjangoModelFactory):
    team = factory.SubFactory(TeamFactory)
    league = factory.SubFactory("wbcore.contrib.example_app.factories.LeagueFactory")
    points = factory.Faker("pyint")
    match_points_for = factory.Faker("pyint")
    match_points_against = factory.Faker("pyint")
    match_point_difference = factory.Faker("pyint", min_value=-500)
    wins = factory.Faker("pyint")
    draws = factory.Faker("pyint")
    losses = factory.Faker("pyint")
    form = factory.Faker("text", max_nb_chars=5)

    class Meta:
        model = TeamResults



from django.utils.translation import gettext as _
from wbcore.contrib.example_app.models import Stadium
from wbcore.metadata.configs.titles import TitleViewConfig


class TeamTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Teams")

    def get_create_title(self) -> str:
        return _("Create Team")

    def get_instance_title(self) -> str:
        return _("Team")


class TeamStadiumTitleConfig(TeamTitleConfig):
    def get_create_title(self) -> str:
        if stadium_id := self.view.kwargs.get("stadium_id"):
            try:
                return _("Create Team In {}").format(Stadium.objects.get(id=stadium_id).name)
            except Stadium.DoesNotExist:
                pass
        return super().get_create_title()



import pytest
from django.db.utils import IntegrityError
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.notifications.factories.notification_types import (
    NotificationTypeModelFactory,
)
from wbcore.contrib.notifications.models import (
    NotificationType,
    NotificationTypeSetting,
)


@pytest.mark.django_db
class TestNotificationType:
    def test_factory(self, notification_type: NotificationType):
        assert isinstance(notification_type, NotificationType)
        assert notification_type.pk is not None

    def test_to_str(self, notification_type: NotificationType):
        assert str(notification_type) == notification_type.title

    def test_endpoint_basename(self):
        assert NotificationType.get_endpoint_basename() == "wbcore:notifications:notification_type"

    def test_representation_endpoint(self):
        assert (
            NotificationType.get_representation_endpoint()
            == "wbcore:notifications:notification_type_representation-list"
        )

    def test_representation_value_key(self):
        assert NotificationType.get_representation_value_key() == "id"

    def test_representation_label_key(self):
        assert NotificationType.get_representation_label_key() == "{{title}}"

    @pytest.mark.parametrize("notification_type__code", ["code"])
    def test_unique_code(self, notification_type: NotificationType):
        assert notification_type.pk is not None

        with pytest.raises(IntegrityError):
            NotificationTypeModelFactory(code="code")


@pytest.mark.django_db
class TestNotificationTypeSetting:
    def test_factory(self, notification_type_setting: NotificationTypeSetting):
        assert isinstance(notification_type_setting, NotificationTypeSetting)
        assert notification_type_setting.pk is not None

    def test_to_str(self, notification_type_setting: NotificationTypeSetting):
        template = "{0.user}: {0.notification_type} ({0.enable_web}/{0.enable_mobile}/{0.enable_email})"
        assert str(notification_type_setting) == template.format(notification_type_setting)

    def test_endpoint_basename(self):
        assert NotificationTypeSetting.get_endpoint_basename() == "wbcore:notifications:notification_type_setting"

    def test_representation_endpoint(self):
        assert (
            NotificationTypeSetting.get_representation_endpoint()
            == "wbcore:notifications:notification_type_setting_representation-list"
        )

    def test_representation_value_key(self):
        assert NotificationTypeSetting.get_representation_value_key() == "id"

    def test_representation_label_key(self):
        assert NotificationTypeSetting.get_representation_label_key() == "{{notification_type}}"

    def test_create_through_post_save_from_notification_type(self, user):
        assert user.pk is not None
        type_count = NotificationTypeSetting.objects.filter(user=user).count()

        NotificationTypeModelFactory()

        assert NotificationTypeSetting.objects.filter(user=user).count() == type_count + 1

    def test_create_through_post_save_from_user(self, notification_type):
        assert notification_type.pk is not None
        assert not NotificationTypeSetting.objects.filter(notification_type=notification_type).exists()

        UserFactory()

        assert NotificationTypeSetting.objects.filter(notification_type=notification_type).exists()



from contextlib import suppress
from typing import Generator

from wbcore.contrib.icons import WBIcon
from wbcore.enums import Button
from wbcore.metadata.configs.buttons.buttons import DropDownButton
from wbcore.signals.instance_buttons import (
    add_button,
    add_extra_button,
    add_instance_button,
)
from wbcore.utils.importlib import parse_signal_received_for_module

from ..base import WBCoreViewConfig


class ButtonViewConfig(WBCoreViewConfig):
    metadata_key = "buttons"
    config_class_attribute = "button_config_class"

    def serialize(self, base_buttons: set[Button], remote_resources):
        """
        Serialize the button view config

        Args:
            base_buttons: the base button gathered from the view config
            remote_resources: the return of the send function
        Returns:
            Yield the serialized button, without duplicates and appends the module prefix to the remote button
        """
        base_buttons = list(zip([None] * len(base_buttons), base_buttons))  # append an empty perfix for base buttons
        for prefix, btn in parse_signal_received_for_module(remote_resources):
            base_buttons.append((prefix, btn))

        keys = []
        for key_prefix, btn in sorted(
            filter(lambda x: x[1], base_buttons), key=lambda e: (e[1].weight, e[1].label, e[1].title)
        ):
            btn_serialized = btn.serialize(self.request, key_prefix=key_prefix)
            if "key" not in btn_serialized or btn_serialized["key"] not in keys:  # ensure we don't include duplicates
                yield btn_serialized

    # Utils
    def order_buttons(self, buttons: set, ordering: list):
        yield from filter(lambda b: b in buttons, ordering)

    # FSM Button Configuration
    FSM_LIST = True
    FSM_INSTANCE = True
    FSM_DROPDOWN = False
    FSM_DROPDOWN_ICON = WBIcon.ADD.icon
    FSM_DROPDOWN_LABEL = "Transitions"
    FSM_WEIGHT = 100

    def get_fsm_buttons(self) -> set:
        if self.FSM_DROPDOWN and (FSM_BUTTONS := getattr(self.view, "FSM_BUTTONS")) and len(FSM_BUTTONS) > 0:
            return {
                DropDownButton(
                    label=self.FSM_DROPDOWN_LABEL,
                    icon=self.FSM_DROPDOWN_ICON,
                    title=self.FSM_DROPDOWN_LABEL,
                    weight=self.FSM_WEIGHT,
                    buttons=tuple(FSM_BUTTONS),
                )
            }
        return getattr(self.view, "FSM_BUTTONS", set())

    # list Button Configuration
    LIST_BUTTONS = frozenset({Button.NEW.value, Button.REFRESH.value})
    LIST_BUTTONS_ORDERING = [Button.NEW.value, Button.REFRESH.value]

    def get_list_buttons(self) -> Generator[None, str, None]:
        if content_type := self.view.get_content_type():
            buttons = set(self.LIST_BUTTONS)
            add_permission = f"{content_type.app_label}.add_{content_type.model}"

            if not self.request.user.has_perm(add_permission):
                with suppress(KeyError):
                    buttons.remove(Button.NEW.value)

            yield from self.order_buttons(buttons, self.LIST_BUTTONS_ORDERING)

    # Instance Button Configuration
    INSTANCE_BUTTONS = frozenset({Button.SAVE.value, Button.REFRESH.value, Button.DELETE.value})
    INSTANCE_BUTTONS_ORDERING = [Button.SAVE.value, Button.REFRESH.value, Button.DELETE.value]

    def get_instance_buttons(self) -> Generator[None, str, None]:
        if content_type := self.view.get_content_type():
            buttons = set(self.INSTANCE_BUTTONS)
            change_permission = f"{content_type.app_label}.change_{content_type.model}"
            delete_permission = f"{content_type.app_label}.delete_{content_type.model}"

            if not self.request.user.has_perm(change_permission):
                with suppress(KeyError):
                    buttons.remove(Button.SAVE.value)

            if not self.request.user.has_perm(delete_permission):
                with suppress(KeyError):
                    buttons.remove(Button.DELETE.value)

            yield from self.order_buttons(buttons, self.INSTANCE_BUTTONS_ORDERING)

    # Create Button Configuration
    CREATE_BUTTONS = frozenset(
        {Button.SAVE.value, Button.SAVE_AND_CLOSE.value, Button.SAVE_AND_NEW.value, Button.RESET.value}
    )
    CREATE_BUTTONS_ORDERING = [
        Button.SAVE.value,
        Button.SAVE_AND_CLOSE.value,
        Button.SAVE_AND_NEW.value,
        Button.RESET.value,
    ]

    def get_create_buttons(self) -> Generator[None, str, None]:
        buttons = set(self.CREATE_BUTTONS)
        yield from self.order_buttons(buttons, self.CREATE_BUTTONS_ORDERING)

    # Custom Instance Button Configuration
    CUSTOM_INSTANCE_BUTTONS = frozenset()
    CUSTOM_LIST_INSTANCE_BUTTONS = frozenset()
    CUSTOM_EXTRA_BUTTONS = frozenset()

    def get_custom_instance_buttons(self) -> set:
        return set(self.CUSTOM_INSTANCE_BUTTONS)

    def get_custom_list_instance_buttons(self) -> set:
        return set(self.CUSTOM_LIST_INSTANCE_BUTTONS)

    def _get_custom_instance_buttons(self) -> Generator:
        custom_instance_buttons = set()
        if self.instance:
            custom_instance_buttons |= self.get_custom_instance_buttons()
            if self.FSM_INSTANCE:
                custom_instance_buttons |= self.get_fsm_buttons()

        else:
            custom_instance_buttons |= self.get_custom_list_instance_buttons()
            if self.FSM_LIST:
                custom_instance_buttons |= self.get_fsm_buttons()

        yield from self.serialize(
            custom_instance_buttons,
            add_instance_button.send(
                self.view.__class__, many=not self.instance, request=self.request, view=self.view, **self.view.kwargs
            ),
        )

    def get_custom_extra_buttons(self) -> set:
        return set(self.CUSTOM_EXTRA_BUTTONS)

    def _get_custom_extra_buttons(self):
        yield from self.serialize(
            self.get_custom_extra_buttons(),
            add_extra_button.send(
                self.view.__class__, instance=self.instance, request=self.request, view=self.view, **self.view.kwargs
            ),
        )

    # Custom Button Configuration
    CUSTOM_BUTTONS = frozenset()

    def get_custom_buttons(self) -> set:
        return set(self.CUSTOM_BUTTONS)

    def _get_custom_buttons(self):
        yield from self.serialize(
            self.get_custom_buttons(),
            add_button.send(self.view.__class__, request=self.request, view=self.view, **self.view.kwargs),
        )

    def get_metadata(self) -> dict:
        return {
            "custom_instance": self._get_custom_instance_buttons(),
            "custom": self._get_custom_buttons(),
            "extra": self._get_custom_extra_buttons(),
        }



import pytest
from rest_framework.exceptions import ValidationError
from wbcore.serializers import ChoiceField
from wbcore.serializers.fields.types import WBCoreType


class TestChoiceField:
    CHOICES = [
        ("choice1", "Choice 1"),
        ("choice2", "Choice 2"),
    ]

    def setup_method(self):
        self.field = ChoiceField(choices=self.CHOICES)

    def test_not_none(self):
        assert self.field is not None

    @pytest.mark.parametrize("choice", [choice for choice, choice_repr in CHOICES])
    def test_to_internal_value(self, choice):
        assert self.field.to_internal_value(choice) == choice

    @pytest.mark.parametrize("choice", [choice for choice, choice_repr in CHOICES])
    def test_to_representation(self, choice):
        assert self.field.to_representation(choice) == choice

    def test_to_internal_value_validation_error(self):
        with pytest.raises(ValidationError):
            self.field.to_internal_value("choice3")

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.SELECT.value

    def test_representation(self):
        choices = [{"label": choice[1], "value": choice[0]} for choice in self.CHOICES]
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "choices": choices,
                "decorators": [],
                "depends_on": [],
            },
        )



import pytest
from django.contrib.auth.models import Permission
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APIClient, APIRequestFactory
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.agenda.viewsets import CalendarItemViewSet
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.test.utils import get_kwargs, get_or_create_superuser

from ..factories import CalendarItemFactory


@pytest.mark.django_db
class TestCalendarItemViewSet:
    api_factory = APIRequestFactory()

    def test_activity_is_draggable(self, calendar_item_factory, user_factory):
        calendar_item = calendar_item_factory()
        request = self.api_factory.get("")
        request.user = user_factory(is_active=True, is_superuser=True)
        view = CalendarItemViewSet.as_view({"get": "retrieve"})
        response = view(request, pk=calendar_item.id).render()
        assert response.data["instance"]["is_draggable"] is False

    # =================================================================================================================
    #                                            TESTING PRIVATE CALENDAR ITEMS
    # =================================================================================================================

    def test_entity_can_see_private_item(self, calendar_item_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        activity = calendar_item_factory(entities=[user.profile], visibility=CalendarItem.Visibility.PRIVATE)
        request = self.api_factory.get("")
        request.user = user
        view = CalendarItemViewSet.as_view({"get": "list"})
        response = view(request).render()

        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == activity.id
        assert response.data["results"][0]["is_private"] is False
        assert response.data["results"][0]["title"] == activity.title

    @pytest.mark.parametrize("mvs", [CalendarItemViewSet])
    def test_random_cannot_see_private_item(self, mvs, calendar_item_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        item = calendar_item_factory(visibility=CalendarItem.Visibility.PRIVATE)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()

        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == item.id
        assert response.data["results"][0]["is_private"] is True
        assert response.data["results"][0]["title"] == "Private CalendarItem"

    # =================================================================================================================
    #                                            TESTING CONFIDENTIAL CALENDAR ITEMS
    # =================================================================================================================

    @pytest.mark.parametrize("mvs", [CalendarItemViewSet])
    def test_manager_can_see_confidential_item(self, mvs, calendar_item_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        permission = Permission.objects.get(
            content_type__app_label="agenda", codename="administrate_confidential_items"
        )
        user.user_permissions.add(permission)
        item = calendar_item_factory(visibility=CalendarItem.Visibility.CONFIDENTIAL)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()

        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == item.id
        assert response.data["results"][0]["is_confidential"] is False
        assert response.data["results"][0]["title"] == item.title

    @pytest.mark.parametrize("mvs", [CalendarItemViewSet])
    def test_random_cannot_see_confidential_item(self, mvs, calendar_item_factory, user_factory):
        user = user_factory(is_active=True, is_superuser=True)
        item = calendar_item_factory(visibility=CalendarItem.Visibility.CONFIDENTIAL)
        request = self.api_factory.get("")
        request.user = user
        view = mvs.as_view({"get": "list"})
        response = view(request).render()

        assert response.status_code == status.HTTP_200_OK
        assert response.data["results"]
        assert len(response.data["results"]) == 1
        assert response.data["results"][0]["id"] == item.id
        assert response.data["results"][0]["is_confidential"] is True
        assert response.data["results"][0]["title"] == "Confidential CalendarItem"


@pytest.mark.django_db
class TestSpecificsInfiniteViewsets:
    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (CalendarItemViewSet, CalendarItemFactory),
        ],
    )
    def test_option_request(self, mvs, factory, entry):
        request = APIRequestFactory().options("")
        request.user = get_or_create_superuser()
        obj = factory(entities=(entry,))
        request.GET = request.GET.copy()
        request.GET["date_gte"] = str(obj.period.lower.date())
        kwargs = {}
        mvs.request = request
        vs = mvs.as_view({"options": "options"})
        kwargs = get_kwargs(obj, mvs, request=request)
        response = vs(request, **kwargs).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data

    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (CalendarItemViewSet, CalendarItemFactory),
        ],
    )
    def test_viewsets(self, mvs, factory, entry):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        obj = factory(entities=(entry,))
        request.GET = request.GET.copy()
        request.GET["date_gte"] = str(obj.period.lower.date())
        kwargs = {}
        mvs.request = request
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data

    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (CalendarItemViewSet, CalendarItemFactory),
        ],
    )
    def test_viewsets_without_date_gte(self, mvs, factory, entry):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        factory(entities=(entry,))
        request.GET = request.GET.copy()
        request.GET["date_gte"] = None
        kwargs = {}
        mvs.request = request
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data

    def test_get_ics(self, calendar_item_factory):
        client = APIClient()
        user = UserFactory(is_active=True, is_superuser=True)
        calendar_item_factory(
            entities=(user.profile,),
        )

        client.force_authenticate(user)
        response = client.get(reverse("wbcore:agenda:get_ics", args=[]))
        assert response.status_code == status.HTTP_200_OK



from django.db.models import F
from django.shortcuts import get_object_or_404, render
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from reversion.models import Revision, Version
from reversion_compare.mixins import CompareMixin
from wbcore import viewsets

from ..filters import RevisionFilterSet, VersionFilterSet
from ..serializers import (
    RevisionModelSerializer,
    RevisionRepresentationSerializer,
    VersionModelSerializer,
    VersionRepresentationSerializer,
)
from .buttons import VersionButtonConfig
from .displays import RevisionDisplayConfig, VersionDisplayConfig
from .endpoints import (
    RevisionEndpointConfig,
    VersionEndpointConfig,
    VersionRevisionEndpointConfig,
)
from .titles import VersionTitleConfig


class VersionRepresentationViewSet(viewsets.RepresentationViewSet):
    ordering_fields = ordering = ("id",)
    # search_fields = ('user',)
    serializer_class = VersionRepresentationSerializer
    queryset = Version.objects.select_related("revision")
    filterset_class = VersionFilterSet


class RevisionRepresentationViewSet(viewsets.RepresentationViewSet):
    ordering_fields = ordering = ("id",)
    search_fields = ("user",)
    serializer_class = RevisionRepresentationSerializer
    queryset = Revision.objects.all()


class RevisionModelViewSet(viewsets.ModelViewSet):
    display_config_class = RevisionDisplayConfig
    endpoint_config_class = RevisionEndpointConfig

    queryset = Revision.objects.select_related("user")
    serializer_class = RevisionModelSerializer

    ordering = "-date_created"
    ordering_fields = ("id", "date_created")
    filterset_class = RevisionFilterSet


class VersionModelViewSet(CompareMixin, viewsets.ModelViewSet):
    display_config_class = VersionDisplayConfig
    endpoint_config_class = VersionEndpointConfig
    title_config_class = VersionTitleConfig
    button_config_class = VersionButtonConfig
    queryset = Version.objects.select_related("revision")
    serializer_class = VersionModelSerializer
    filterset_class = VersionFilterSet

    ordering_fields = ("date_created",)
    ordering = ("-date_created",)

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                date_created=F("revision__date_created"), profile_repr=F("revision__user__profile__computed_str")
            )
        )

    @action(detail=True, methods=["GET"], permission_classes=[IsAuthenticated])
    def revert(self, request, pk=None):
        if request.user.is_superuser or request.user.has_perm("reversion.change_version"):
            version1 = get_object_or_404(Version, pk=pk)
            version1.revert()
        return Response({}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=True, methods=["GET"], permission_classes=[IsAuthenticated])
    def comparewith(self, request, pk=None):
        if compare_with := request.GET.get("compare_with", None):
            version1 = get_object_or_404(Version, pk=pk)
            version2 = get_object_or_404(Version, pk=compare_with)
            if version1.id > version2.id:
                # Compare always the newest one (#2) with the older one (#1)
                version1, version2 = version2, version1

            obj = version1.content_type.model_class().objects.get(id=version1.object_id)
            compare_data, has_unfollowed_fields = self.compare(obj, version1, version2)
            return render(
                request,
                "reversion/compare_detail.html",
                {
                    "compare_data": compare_data,
                    "has_unfollowed_fields": has_unfollowed_fields,
                    "version1": version1,
                    "version2": version2,
                },
            )
        return Response({}, status=status.HTTP_400_BAD_REQUEST)


class VersionRevisionModelViewSet(VersionModelViewSet):
    endpoint_config_class = VersionRevisionEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(revision=self.kwargs["revision_id"])



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from .viewsets import GeographyModelViewSet, GeographyRepresentationViewSet

router = WBCoreRouter()
router.register(r"geographyrepresentation", GeographyRepresentationViewSet, basename="geographyrepresentation")
router.register(r"geography", GeographyModelViewSet, basename="geography")

urlpatterns = [
    path("", include(router.urls)),
]



import pytest
from django_fsm import TransitionNotAllowed
from dynamic_preferences.registries import global_preferences_registry
from wbcore.contrib.directory.models import ClientManagerRelationship as CMR
from wbcore.contrib.directory.models import Company, Person
from wbcore.contrib.directory.models.entries import handle_user_deactivation

from ..factories import ClientManagerRelationshipFactory as CMRF


@pytest.mark.django_db
class TestSpecificModelsContacts:
    def test_set_entry_primary_telephone(self, telephone_contact_factory):
        tc = telephone_contact_factory()
        tc2 = telephone_contact_factory()

        tc.set_entry_primary_telephone(tc.entry, tc2.number)
        assert str(tc.entry.primary_telephone_contact()) == str(tc2.entry.primary_telephone_contact())

        tc3 = telephone_contact_factory(primary=False)
        tc3.primary = False
        tc3.save()
        tc3.set_entry_primary_telephone(tc3.entry, tc2.number)
        assert str(tc3.entry.primary_telephone_contact()) == str(tc2.entry.primary_telephone_contact())

    def test_set_entry_primary_email(self, email_contact_factory):
        ec = email_contact_factory()
        ec2 = email_contact_factory()

        ec.set_entry_primary_email(ec.entry, ec2.address)
        assert ec.entry.primary_email_contact().id != ec2.entry.primary_email_contact().id
        assert ec.entry.primary_email_contact().address == ec2.entry.primary_email_contact().address

        ec3 = email_contact_factory(primary=False)
        ec3.primary = False
        ec3.save()
        ec3.set_entry_primary_email(ec3.entry, ec2.address)
        assert ec3.entry.primary_email_contact().id != ec2.entry.primary_email_contact().id
        assert ec3.entry.primary_email_contact().address == ec2.entry.primary_email_contact().address


@pytest.mark.django_db
class TestSpecificModelsEntries:
    def test_primary_email_contact(self, person_factory, company_factory):
        assert company_factory().primary_email_contact() is None
        assert person_factory().primary_email_contact() is None

    def test_primary_telephone_contact(self, person_factory, company_factory):
        assert company_factory().primary_telephone_contact() is None
        assert person_factory().primary_telephone_contact() is None

    def test_primary_address_contact(self, person_factory, company_factory):
        assert company_factory().primary_address_contact() is None
        assert person_factory().primary_address_contact() is None

    def test_primary_website_contact(self, person_factory, company_factory):
        assert company_factory().primary_website_contact() is None
        assert person_factory().primary_website_contact() is None

    def test_primary_banking_contact(self, person_factory, company_factory):
        assert company_factory().primary_banking_contact() is None
        assert person_factory().primary_banking_contact() is None

    def test_get_casted_entry(self, person_factory, company_factory):
        assert company_factory().get_casted_entry()
        assert person_factory().get_casted_entry()

    def test_delete_additional_fields(self, person_factory, company_factory):
        person1 = person_factory()
        person1.additional_fields["special_key"] = "additionnal fields special key"
        person1.delete_additional_fields("special_key")
        assert "special_key" not in person1.additional_fields.keys()

    def test_full_name(self, person_factory, company_factory, internal_user_factory):
        assert person_factory().full_name
        person = internal_user_factory().profile
        person.employers.add(company_factory())
        assert person.str_full()

    # Tests for CRM user deactivation method
    def test_main_company_removed_from_deactivated_user(self, internal_user_factory):
        old_person = internal_user_factory().profile
        main_company_id = global_preferences_registry.manager()["directory__main_company"]
        main_company = Company.objects.get(id=main_company_id)
        assert (
            handle_user_deactivation(sender=None, instance=old_person, substitute_profile=None)[0]
            == f"Removed {main_company.computed_str} from {old_person.computed_str}'s employers"
        )
        assert main_company not in old_person.employers.all()

    def test_no_substitute_person(self, client_manager_relationship_factory, internal_user_factory):
        person = internal_user_factory().profile
        relationship = client_manager_relationship_factory(relationship_manager=person)
        handle_user_deactivation(sender=None, instance=person, substitute_profile=None)
        assert CMR.objects.get(id=relationship.id).status == CMR.Status.REMOVED

    def test_not_approved_substitute_relationships_exists(
        self, client_manager_relationship_factory, person_factory, internal_user_factory
    ):
        old_person = internal_user_factory().profile
        new_person = person_factory()
        old_relationship = client_manager_relationship_factory(
            relationship_manager=old_person, status=CMR.Status.PENDINGADD
        )
        substitute_relationship = client_manager_relationship_factory(
            client=old_relationship.client, relationship_manager=new_person
        )
        assert (
            handle_user_deactivation(sender=None, instance=old_person, substitute_profile=new_person)[1]
            == f"Assigned 1 manager role(s) to {new_person.computed_str}"
        )
        assert not CMR.objects.filter(id=old_relationship.id).exists()
        assert CMR.objects.get(id=substitute_relationship.id).client.id == old_relationship.client.id
        assert CMR.objects.get(id=substitute_relationship.id).relationship_manager.id == new_person.id

    def test_not_approved_substitute_relationships_missing(
        self, client_manager_relationship_factory, person_factory, internal_user_factory
    ):
        old_person = internal_user_factory().profile
        new_person = person_factory()
        old_relationship = client_manager_relationship_factory(
            relationship_manager=old_person, status=CMR.Status.PENDINGADD
        )
        assert (
            handle_user_deactivation(sender=None, instance=old_person, substitute_profile=new_person)[1]
            == f"Assigned 1 manager role(s) to {new_person.computed_str}"
        )
        assert CMR.objects.get(id=old_relationship.id).relationship_manager.id == new_person.id
        assert CMR.objects.get(id=old_relationship.id).client.id == old_relationship.client.id

    def test_approved_with_substitute_relationships(
        self, client_manager_relationship_factory, person_factory, internal_user_factory
    ):
        old_person = internal_user_factory().profile
        new_person = person_factory()
        old_relationship = client_manager_relationship_factory(relationship_manager=old_person, primary=True)
        client_manager_relationship_factory(relationship_manager=old_person)
        substitute_relationship = client_manager_relationship_factory(
            client=old_relationship.client, relationship_manager=new_person, status=CMR.Status.PENDINGADD
        )
        assert (
            handle_user_deactivation(sender=None, instance=old_person, substitute_profile=new_person)[1]
            == f"Assigned 2 manager role(s) to {new_person.computed_str}"
        )
        assert CMR.objects.get(id=substitute_relationship.id).client.id == old_relationship.client.id
        assert CMR.objects.get(id=substitute_relationship.id).relationship_manager.id == new_person.id
        assert CMR.objects.get(id=substitute_relationship.id).status == CMR.Status.APPROVED
        assert CMR.objects.get(id=substitute_relationship.id).primary is True

    def test_approved_without_substitute_relationships_needs_primary(
        self, client_manager_relationship_factory, person_factory, internal_user_factory
    ):
        old_person = internal_user_factory().profile
        new_person = person_factory()
        old_relationship = client_manager_relationship_factory(relationship_manager=old_person, primary=True)
        assert (
            handle_user_deactivation(sender=None, instance=old_person, substitute_profile=new_person)[1]
            == f"Assigned 1 manager role(s) to {new_person.computed_str}"
        )
        assert CMR.objects.filter(
            relationship_manager=new_person, client=old_relationship.client, primary=True, status=CMR.Status.APPROVED
        ).exists()
        assert CMR.objects.get(id=old_relationship.id).status == CMR.Status.REMOVED

    def test_approved_without_substitute_relationships_doesnt_need_primary(
        self, client_manager_relationship_factory, person_factory, internal_user_factory
    ):
        old_person = internal_user_factory().profile
        new_person = person_factory()
        old_relationship = client_manager_relationship_factory(relationship_manager=old_person)
        client_manager_relationship_factory(client=old_relationship.client, primary=True)
        assert (
            handle_user_deactivation(sender=None, instance=old_person, substitute_profile=new_person)[1]
            == f"Assigned 1 manager role(s) to {new_person.computed_str}"
        )
        assert CMR.objects.filter(
            relationship_manager=new_person, client=old_relationship.client, primary=False, status=CMR.Status.APPROVED
        ).exists()
        assert CMR.objects.get(id=old_relationship.id).status == CMR.Status.REMOVED

    def test_soft_deleted_entry_dont_show_in_queryset(self, company_factory, person_factory):
        "ensure the default queryset filter out soft deleted entries"
        person = person_factory.create()
        assert set(Person.objects.all()) == {person}

        person.delete()
        assert not Person.objects.exists()

        company = company_factory.create()
        assert set(Company.objects.all()) == {company}

        company.delete()
        assert not Company.objects.exists()

    def test_soft_deleted_person_doesnt_crash_user_registration(self, user_factory, email_contact_factory):
        user = user_factory(profile=None)
        email_contact_factory(address=user.email, entry__is_active=False)
        assert Person.get_or_create_with_user(user)


@pytest.mark.django_db
class TestSpecificModelsClientManagerRelationship:
    @pytest.mark.parametrize("status, expected", [("DRAFT", "PENDINGADD")])
    def test_submit(self, status, expected):
        cmr = CMRF(status=CMR.Status.DRAFT)
        cmr.submit()
        assert cmr.status == expected

    @pytest.mark.parametrize("status, expected", [("PENDINGADD", "APPROVED")])
    def test_approve(self, status, expected):
        cmr = CMRF(status=CMR.Status.PENDINGADD)
        cmr.approve()
        assert cmr.status == expected

    @pytest.mark.parametrize("status, expected", [("PENDINGADD", "DRAFT")])
    def test_deny(self, status, expected):
        cmr = CMRF(status=CMR.Status.PENDINGADD)
        cmr.deny()
        assert cmr.status == expected

    @pytest.mark.parametrize("status, expected", [("DRAFT", "APPROVED")])
    def test_mngapprove(self, status, expected):
        cmr = CMRF(status=CMR.Status.DRAFT)
        cmr.mngapprove()
        assert cmr.status == expected

    @pytest.mark.parametrize("status, expected", [("PENDINGREMOVE", "APPROVED")])
    def test_denyremoval(self, status, expected):
        cmr = CMRF(status=CMR.Status.PENDINGREMOVE)
        cmr.denyremoval()
        assert cmr.status == expected

    @pytest.mark.parametrize("status, expected", [("PENDINGREMOVE", "REMOVED")])
    def test_approveremoval(self, status, expected):
        cmr = CMRF(status=CMR.Status.PENDINGREMOVE)
        cmr.approveremoval()
        assert cmr.status == expected

    @pytest.mark.parametrize("status, expected", [("APPROVED", "PENDINGADD")])
    def test_makeprimary(self, status, expected):
        cmr1 = CMRF(primary=True)
        cmr2 = CMRF(primary=False, client=cmr1.client)
        cmr2.makeprimary()
        assert cmr2.status == expected
        assert cmr2.primary
        with pytest.raises(TransitionNotAllowed):
            cmr1.makeprimary()
        assert cmr1.status == status
        assert cmr1.primary

    @pytest.mark.parametrize("status, expected", [("APPROVED", "PENDINGREMOVE")])
    def test_remove(self, status, person, expected):
        cmr1 = CMRF.create(primary=True, client=person)
        cmr2 = CMRF.create(primary=False, client=person)
        cmr2.remove()
        assert cmr2.status == expected
        with pytest.raises(TransitionNotAllowed):
            cmr1.remove()
        assert cmr1.status == status

    @pytest.mark.parametrize("status, expected", [("REMOVED", "PENDINGADD")])
    def test_reinstate(self, status, expected):
        cmr = CMRF(status=CMR.Status.REMOVED)
        cmr.reinstate()
        assert cmr.status == expected


@pytest.mark.django_db
class TestSpecificModelsRelationships:
    def test_relationship_type_str(self, relationship_type_factory):
        assert relationship_type_factory(title="Type").__str__() == "Type"

    def test_relationship_str(self, relationship_factory, relationship_type_factory, person_factory):
        relationship_type = relationship_type_factory(
            title="Type",
            counter_relationship=None,
        )
        from_entry = person_factory(first_name="John", last_name="Doe")
        to_entry = person_factory(first_name="Jane", last_name="Doe")
        rel = relationship_factory(
            relationship_type=relationship_type,
            from_entry=from_entry,
            to_entry=to_entry,
        )
        assert rel.__str__() == "John Doe is Type of Jane Doe"



import re

from django.db import transaction
from rest_framework.serializers import ValidationError
from wbcore import serializers as wb_serializers
from wbcore.serializers import ModelSerializer, RepresentationSerializer

from .models import (
    DataBackend,
    ImportSource,
    ParserHandler,
    Source,
    import_data_as_task,
)


class ImportSourceRepresentationSerializer(RepresentationSerializer):
    class Meta:
        model = ImportSource
        fields = ("id", "file")


class SourceRepresentationSerializer(RepresentationSerializer):
    def __init__(self, *args, label_key="{{id}} - {{title}}", **kwargs):
        self.allowed_sources = kwargs.pop("allowed_sources", Source.objects.none())
        super().__init__(*args, label_key=label_key, **kwargs)

    def get_filter_params(self, request):
        filter_params = {}
        if self.allowed_sources.exists():
            filter_params["id__in"] = ",".join(map(str, self.allowed_sources.values_list("id", flat=True)))
        return filter_params

    class Meta:
        model = Source
        fields = ("id", "title")


class ParserHandlerRepresentationSerializer(RepresentationSerializer):
    def __init__(self, *args, **kwargs):
        self.allowed_handler = kwargs.pop("allowed_handler", None)
        super().__init__(*args, **kwargs)

    def get_filter_params(self, request):
        filter_params = {}
        if self.allowed_handler:
            filter_params["handler__icontains"] = self.allowed_handler
        return filter_params

    class Meta:
        model = ParserHandler
        fields = ("id", "parser", "handler")


class DataBackendRepresentationSerializer(RepresentationSerializer):
    def __init__(self, *args, **kwargs):
        self.allowed_backend_class_path = kwargs.pop("allowed_backend_class_path", ParserHandler.objects.none())
        super().__init__(*args, **kwargs)

    def get_filter_params(self, request):
        filter_params = {}
        if self.allowed_backend_class_path:
            filter_params["backend_class_path__icontains"] = self.allowed_backend_class_path
        return filter_params

    class Meta:
        model = DataBackend
        fields = ("id", "title", "backend_class_path")


class ParserHandlerModelSerializer(ModelSerializer):
    class Meta:
        model = ParserHandler
        fields = ("id", "parser", "handler")


class ImportSourceModelSerializer(ModelSerializer):
    parser_handler = wb_serializers.CharField(required=False)

    def validate(self, data):
        if "creator" not in data and "request" in self.context:
            data["creator"] = self.context["request"].user
        if parser_handler := data.get("parser_handler"):
            if match := re.search(r"([\w\.]+)::(\w+\.\w+)", parser_handler):
                parser_lookup = dict(parser=match.group(1), handler=match.group(2))
            else:
                parser_lookup = dict(id=parser_handler)
            try:
                data["parser_handler"] = ParserHandler.objects.get(**parser_lookup)
            except (ParserHandler.DoesNotExist, ValueError):
                raise ValidationError({"parser_handler": "Invalid parser handler"})
        return data

    def create(self, validated_data):
        if "parser_handler" not in validated_data and "source" in validated_data:
            # if parser handler is not provided, we default it to the first source parser handler
            validated_data["parser_handler"] = validated_data["source"].parser_handler.first()
        instance = super().create(validated_data)
        transaction.on_commit(import_data_as_task.s(instance.id).delay)
        return instance

    class Meta:
        model = ImportSource
        read_only_fields = ("id", "status")
        fields = (
            "id",
            "status",
            "origin",
            "creator",
            "file",
            "save_data",
            "parser_handler",
            "source",
        )


class SourceModelSerializer(wb_serializers.ModelSerializer):
    _parser_handler = ParserHandlerRepresentationSerializer(source="parser_handler", many=True)
    _data_backend = DataBackendRepresentationSerializer(source="data_backend")
    crontab = wb_serializers.StringRelatedField()

    class Meta:
        model = Source
        fields = (
            "title",
            "uuid",
            "parser_handler",
            "_parser_handler",
            "is_active",
            "connection_parameters",
            "import_parameters",
            "data_backend",
            "_data_backend",
            "crontab",
            "import_timedelta_interval",
        )



from importlib import import_module
from os import environ

from django.core.management import call_command
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Bootstrap the current module"

    def handle(self, *args, **options):
        self.stdout.write("Migrating Database")
        call_command("migrate")
        settings = import_module(str(environ.get("DJANGO_SETTINGS_MODULE")))
        self.stdout.write("Loading Fixtures...")
        for fixture in getattr(settings, "BOOTSTRAP_FIXTURES", []):
            self.stdout.write(f"Load {fixture}: ", ending="")
            call_command("loaddata", fixture)



from datetime import datetime
from enum import Enum

from django.db.models import Q
from django.forms import ValidationError
from django.utils import timezone
from django.utils.translation import gettext as _
from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.example_app.models import (
    Event,
    EventType,
    League,
    Match,
    Player,
    Role,
    Sport,
    SportPerson,
    Team,
)
from wbcore.contrib.example_app.serializers import (
    LeagueRepresentationSerializer,
    SportPersonRepresentationSerializer,
    SportRepresentationSerializer,
    StadiumRepresentationSerializer,
    TeamRepresentationSerializer,
)
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)


class MatchErrorMessages(Enum):
    same_teams = _("Team cannot play against itself.")
    same_match = _("A match between these teams at this time already exists.")
    wrong_sport = _("League doesn't match the specified sport.")


class EventTypeErrorMessages(Enum):
    same_name = _("An event type with this name already exists for this sport.")


class EventErrorMessages(Enum):
    same_event = _("This event already exists.")
    wrong_duration = _("Event outside of match duration.")
    wrong_event_type = _("Please select a {} event type.")


class MatchModelSerializer(serializers.ModelSerializer):
    _home = TeamRepresentationSerializer(
        source="home",
        optional_get_parameters={"away": "opponent"},
    )
    _away = TeamRepresentationSerializer(
        source="away",
        optional_get_parameters={"home": "opponent"},
    )
    _league = LeagueRepresentationSerializer(
        source="league",
        optional_get_parameters={"sport": "sport"},
        depends_on=[{"field": "sport", "options": {}}],
    )
    _stadium = StadiumRepresentationSerializer(source="stadium")
    _referee = SportPersonRepresentationSerializer(
        source="referee",
        filter_params=lambda request: {"roles": (Role.objects.get(title="Referee").pk)}
        if Role.objects.filter(title="Referee").exists()
        else {},
    )

    _sport = SportRepresentationSerializer(source="sport")

    def validate(self, data: dict) -> dict:
        home: Team | None = data.get("home", self.instance.home if self.instance else None)
        away: Team | None = data.get("away", self.instance.away if self.instance else None)
        date_time: datetime | None = data.get("date_time", self.instance.date_time if self.instance else None)
        sport: Sport | None = data.get("sport", self.instance.sport if self.instance else None)
        if "league" in data:
            league: League = data["league"]
        else:
            league: League | None = self.instance.league if self.instance else None

        if home and away:
            if home.pk == away.pk:
                raise ValidationError({"away": MatchErrorMessages.same_teams.value})
            if date_time:
                match = Match.objects.filter(away=away, date_time=date_time, home=home)
                if obj := self.instance:
                    match = match.exclude(id=obj.pk)
                if match.exists():
                    raise ValidationError({"home": MatchErrorMessages.same_match.value})

        if league and sport:
            if league.sport.pk != sport.pk:
                raise ValidationError({"league": MatchErrorMessages.wrong_sport.value})
        return super().validate(data)

    @serializers.register_dynamic_button()
    def event_type_buttons(self, instance: Match, request, user) -> list:
        def get_current_match_minute() -> int | None:
            """Returns the current match minute based on the match's start time if inside the match duration boundaries

            Returns:
                int|None: The current minute
            """
            minutes_since_match_start = (timezone.now() - instance.date_time).total_seconds() / 60
            if 0 < minutes_since_match_start < instance.sport.match_duration:
                return int(minutes_since_match_start)
            return None

        class EventButtonSerializer(serializers.Serializer):
            _person = SportPersonRepresentationSerializer(source="person", filter_params={"match": instance.pk})
            person = serializers.PrimaryKeyRelatedField(
                label=_("Person"),
                queryset=SportPerson.objects.filter(
                    Q(coached_team__home_matches=instance)
                    | Q(coached_team__away_matches=instance)
                    | Q(refereed_matches=instance)
                    | Q(
                        id__in=Player.objects.filter(
                            Q(current_team__home_matches=instance) | Q(current_team__away_matches=instance)
                        ).values("id")
                    )
                ).distinct(),
            )
            minute = serializers.IntegerField(label=_("Minute"), default=get_current_match_minute)

            def validate(self, data: dict) -> dict:
                match: Match | None = data.get("match", self.instance.match if self.instance else None)
                minute: int | None = data.get("minute", self.instance.minute if self.instance else None)

                if match and minute:
                    if (duration := match.sport.match_duration) < minute:
                        raise ValidationError(
                            {"minute": _("Event outside of match duration ({} min).").format(duration)}
                        )

                return super().validate(data)

        buttons = []
        if instance.status == Match.MatchStatus.ONGOING:
            for event_type in instance.sport.event_types.all():
                button = ActionButton(
                    method=RequestType.PATCH,
                    identifiers=("example_app:event-match",),
                    endpoint=f"{reverse('example_app:event-matchevent', args=[], request=request)}?event_type={event_type.pk}&match={instance.pk}",
                    action_label=event_type.name,
                    title=event_type.name,
                    label=event_type.name,
                    icon=event_type.icon,
                    description_fields="",
                    serializer=EventButtonSerializer,
                    instance_display=create_simple_display([["person"], ["minute"]]),
                )
                button.request = request

                buttons.append(button)
        return buttons

    @serializers.register_only_instance_resource()
    def events_inline(self, instance: Match, request, user, view, **kwargs) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "event_list": reverse("example_app:event-match-list", args=[instance.pk], request=self.context["request"])
        }

    class Meta:
        model = Match
        fields = (
            "id",
            "home",
            "_home",
            "away",
            "_away",
            "sport",
            "_sport",
            "league",
            "_league",
            "date_time",
            "referee",
            "_referee",
            "stadium",
            "_stadium",
            "status",
            "score_home",
            "score_away",
            "computed_str",
            "_additional_resources",
            "_buttons",
        )


class ReadOnlyMatchModelSerializer(MatchModelSerializer):
    class Meta(MatchModelSerializer.Meta):
        read_only_fields = MatchModelSerializer.Meta.fields


class MatchRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:match-detail")

    class Meta:
        model = Match
        fields = ("id", "computed_str", "_detail")


class EventTypeModelSerializer(serializers.ModelSerializer):
    _sport = SportRepresentationSerializer(source="sport")
    color = serializers.ColorPickerField(required=False)

    def validate(self, data: dict) -> dict:
        name: str | None = data.get("name", self.instance.name if self.instance else None)
        sport: Sport | None = data.get("sport", self.instance.sport if self.instance else None)

        if name and sport:
            event_type = EventType.objects.filter(name=name, sport=sport)
            if obj := self.instance:
                event_type = event_type.exclude(id=obj.pk)
            if event_type.exists():
                raise ValidationError({"name": EventTypeErrorMessages.same_name.value})
        return super().validate(data)

    class Meta:
        model = EventType
        fields = "__all__"


class EventTypeRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:eventtype-detail")

    class Meta:
        model = EventType
        fields = ("id", "name", "_detail")


class EventModelSerializer(serializers.ModelSerializer):
    _person = SportPersonRepresentationSerializer(source="person")
    _match = MatchRepresentationSerializer(source="match")
    _event_type = EventTypeRepresentationSerializer(source="event_type")
    event_description = serializers.TextAreaField(required=False)

    def validate(self, data: dict) -> dict:
        person: SportPerson | None = data.get("person", self.instance.person if self.instance else None)
        match: Match | None = data.get("match", self.instance.match if self.instance else None)
        minute: int | None = data.get("minute", self.instance.minute if self.instance else None)
        event_type: EventType | None = data.get("event_type", self.instance.event_type if self.instance else None)

        if match:
            if minute:
                if match.sport.match_duration < minute:
                    raise ValidationError({"minute": EventErrorMessages.wrong_duration.value})
                if person and event_type:
                    event = Event.objects.filter(minute=minute, match=match, person=person, event_type=event_type)
                    if obj := self.instance:
                        event = event.exclude(id=obj.pk)
                    if event.exists():
                        raise ValidationError({"non_field_errors": EventErrorMessages.same_event.value})
            if event_type and match.sport.pk != event_type.sport.pk:
                raise ValidationError(
                    {"event_type": EventErrorMessages.wrong_event_type.value.format(match.sport.name)}
                )

        return super().validate(data)

    class Meta:
        model = Event
        fields = "__all__"


class LeaguePlayerStatisticsModelSerializer(serializers.ModelSerializer):
    person_id = serializers.IntegerField()
    person_name = serializers.CharField()
    count = serializers.IntegerField()
    id = serializers.PrimaryKeyCharField()

    class Meta:
        model = Event
        fields = ("id", "person_id", "person_name", "count")


class LeagueTeamStatisticsModelSerializer(serializers.ModelSerializer):
    team_id = serializers.IntegerField()
    team_name = serializers.CharField()
    count = serializers.IntegerField()
    id = serializers.PrimaryKeyCharField()

    class Meta:
        model = Event
        fields = (
            "id",
            "team_id",
            "count",
            "team_name",
        )


class EventRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:event-detail")

    class Meta:
        model = Event
        fields = ("id", "event_type", "minute", "person", "_detail")



import calendar as calendar_reference
from datetime import date

from dateutil import rrule


def convert_rrulestr_to_dict(
    rrule_str: str, dtstart: date | None = None, count: int | None = None, until: date | None = None
) -> dict:
    kwargs = {
        values[0]: values[1]
        for elt in rrule_str.replace("RRULE:", "").split(";")
        if (values := elt.split("=")) and len(values) == 2
    }
    rule_dict = {}
    if freq := kwargs.get("FREQ"):
        rule_dict = {
            "freq": getattr(rrule, freq),
            "interval": int(kwargs.get("INTERVAL", "1")),
        }
        if dtstart:
            rule_dict["dtstart"] = dtstart
        if count:
            rule_dict["count"] = count
        if until:
            rule_dict["until"] = until
        if wkst := kwargs.get("WKST"):
            rule_dict["wkst"] = getattr(rrule, wkst)
        if byday := kwargs.get("BYDAY"):
            rule_dict["byweekday"] = [getattr(rrule, day.strip()) for day in byday.split(",")]
    return rule_dict


def convert_weekday_rrule_to_day_name(wkday: rrule.weekday) -> str:
    week_dict = {_day: _day.weekday for _day in rrule.weekdays}
    if idx := week_dict.get(wkday):
        return calendar_reference.day_name[idx]



import pytest
from wbcore.metadata.configs.display.instance_display import operators


@pytest.mark.parametrize(
    "method, value, result",
    [
        (operators.lt, 20, "<20"),
        (operators.lte, 20, "<=20"),
        (operators.gt, 20, ">20"),
        (operators.gte, 20, ">=20"),
    ],
)
def test_operator(method, value, result):
    assert method(value) == result



from django.utils.translation import gettext as _
from wbcore.contrib.directory.models import ClientManagerRelationship as CMR
from wbcore.contrib.directory.models import Company, Entry, Person
from wbcore.metadata.configs.titles import TitleViewConfig


class RelationshipEntryModelTitleConfig(TitleViewConfig):
    def get_list_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("Relationships of {entry}").format(entry=str(entry))

    def get_create_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("New Relationship for {entry}").format(entry=str(entry))


class RelationshipTypeModelTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Relationship Types")

    def get_instance_title(self):
        return _("Relationship Type")

    def get_create_title(self):
        return _("New Relationship Type")


class ClientManagerTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        if pk := self.view.kwargs.get("pk"):
            cmr_request = CMR.objects.get(id=pk)
            return _("Client Manager Relationship Between {client} & {person}").format(
                client=str(cmr_request.client), person=str(cmr_request.relationship_manager)
            )

    def get_list_title(self):
        if client_id := self.request.GET.get("client"):
            client = Entry.all_objects.get(id=client_id)
            return _("Relationship Managers for {client}").format(client=str(client))

        elif relationship_manager_id := self.request.GET.get("relationship_manager"):
            relationship_manager = Person.all_objects.get(id=relationship_manager_id)
            return _("Clients of {relationship_manager}").format(relationship_manager=str(relationship_manager))

        return _("Client Manager Relationships")

    def get_create_title(self):
        if client_id := self.request.GET.get("client"):
            client = Entry.all_objects.get(id=client_id)
            return _("New Client Relationship for {client}").format(client=str(client))

        elif relationship_manager_id := self.request.GET.get("relationship_manager"):
            relationship_manager = Person.all_objects.get(id=relationship_manager_id)
            return _("New Manager Relationship for {relationship_manager}").format(
                relationship_manager=str(relationship_manager)
            )

        return _("New Client Manager Relationship")


class EmployerEmployeeRelationshipTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str | None:
        if employee_id := self.view.kwargs.get("employee_id"):
            person = Person.all_objects.get(id=employee_id)
            return f"Employers of {str(person)}"
        if employer_id := self.view.kwargs.get("employer_id"):
            company = Company.all_objects.get(id=employer_id)
            return f"Employees of {str(company)}"
        return super().get_list_title()

    def get_instance_title(self) -> str:
        return _("Employer Employee Relationship")

    def get_create_title(self) -> str | None:
        if employee_id := self.view.kwargs.get("employee_id"):
            person = Person.all_objects.get(id=employee_id)
            return f"New Employer for {str(person)}"
        if employer_id := self.view.kwargs.get("employer_id"):
            company = Company.all_objects.get(id=employer_id)
            return f"New Employee for {str(company)}"
        return super().get_create_title()


class UserIsClientTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        return _("Your Contacts")



import logging
from inspect import getmembers

from rest_framework import serializers
from rest_framework.reverse import reverse
from wbcore.metadata.utils import prefix_key
from wbcore.signals import (
    add_additional_resource,
    add_dynamic_button,
    add_instance_additional_resource,
)
from wbcore.utils.importlib import parse_signal_received_for_module

from .mixins import WBCoreSerializerFieldMixin
from .text import TextField
from .types import WBCoreType

logger = logging.getLogger(__name__)


def register_resource():
    def decorator(func):
        func._is_additional_resource = True
        return func

    return decorator


def register_only_instance_resource():
    def decorator(func):
        func._is_instance_additional_resource = True
        return func

    return decorator


def _is_additional_resource(attr):
    return hasattr(attr, "_is_additional_resource")


def _is_instance_additional_resource(attr):
    return hasattr(attr, "_is_instance_additional_resource")


def register_dynamic_button():
    def decorator(func):
        func._is_dynamic_button = True
        return func

    return decorator


def register_only_instance_dynamic_button():
    def decorator(func):
        func._is_instance_dynamic_button = True
        return func

    return decorator


def _is_dynamic_button(attr):
    return hasattr(attr, "_is_dynamic_button")


def _is_instance_dynamic_button(attr):
    return hasattr(attr, "_is_instance_dynamic_button")


class DynamicButtonField(WBCoreSerializerFieldMixin, serializers.ReadOnlyField):
    field_type = "_dynamic_buttons"

    def get_attribute(self, instance):
        return instance

    def to_representation(self, value):
        buttons = list()
        if request := self.parent.context.get("request"):
            for _, button_func in getmembers(self.parent.__class__, _is_dynamic_button):
                btns = button_func(self.parent, instance=value, request=request, user=request.user)
                buttons.extend([btn.serialize(request) for btn in btns])
            dynamic_buttons = parse_signal_received_for_module(
                add_dynamic_button.send(
                    sender=self.parent.__class__,
                    serializer=self.parent,
                    instance=value,
                    request=request,
                    user=request.user,
                    view=self.parent.context.get("view", None),
                )
            )
            for prefix, btns in dynamic_buttons:
                for btn in btns:
                    setattr(btn, "prefix_key", prefix)
                    buttons.append(btn.serialize(request))
            if (view := self.parent.context.get("view", None)) and not (getattr(view, "action", "list") == "list"):
                for _, button_func in getmembers(self.parent.__class__, _is_instance_dynamic_button):
                    btns = button_func(self.parent, instance=value, request=request, user=request.user)
                    buttons.extend([btn.serialize(request) for btn in btns])

        return buttons


class AdditionalResourcesField(WBCoreSerializerFieldMixin, serializers.ReadOnlyField):
    field_type = "_additional_resources"

    def get_attribute(self, instance):
        return instance

    @property
    def serializer_class_for_remote_additional_resources(self):
        return getattr(self.parent, "SERIALIZER_CLASS_FOR_REMOTE_ADDITIONAL_RESOURCES", self.parent.__class__)

    def to_representation(self, value):
        resources = dict()
        if (request := self.parent.context.get("request")) and (view := self.parent.context.get("view", None)):
            for _, function in getmembers(self.parent.__class__, _is_additional_resource):
                _d = function(self.parent, instance=value, request=request, user=request.user)
                resources.update(_d)

            is_list = getattr(view, "action", "list") == "list"
            remote_resources = parse_signal_received_for_module(
                add_additional_resource.send(
                    sender=self.serializer_class_for_remote_additional_resources,
                    serializer=self.parent,
                    instance=value,
                    request=request,
                    user=request.user,
                    view=view,
                    is_list=is_list,
                )
            )
            for prefix, response in remote_resources:
                for key, remote_resource in response.items():
                    resources[prefix_key(key, prefix)] = remote_resource
            if not is_list:
                for _, function in getmembers(self.parent.__class__, _is_instance_additional_resource):
                    _d = function(self.parent, instance=value, request=request, user=request.user, view=view)
                    resources.update(_d)
                remote_resources = parse_signal_received_for_module(
                    add_instance_additional_resource.send(
                        sender=self.serializer_class_for_remote_additional_resources,
                        serializer=self.parent,
                        instance=value,
                        request=request,
                        user=request.user,
                        view=view,
                        is_list=False,
                    )
                )
                for prefix, response in remote_resources:
                    for key, remote_resource in response.items():
                        resources[prefix_key(key, prefix)] = remote_resource
        return resources


class HyperlinkField(WBCoreSerializerFieldMixin, serializers.ReadOnlyField):
    def __init__(self, *args, **kwargs):
        self.reverse_name = kwargs.pop("reverse_name")
        self.id_field_name = kwargs.pop("id_field_name", "id")
        super().__init__(*args, **kwargs)

    def get_attribute(self, obj):
        request = self.context.get("request", None)
        if obj:
            obj_id = getattr(obj, self.id_field_name, obj.pk)
            if request:
                return reverse(self.reverse_name, args=[obj_id], request=request)
            return reverse(self.reverse_name, args=[obj_id])


class ReadOnlyField(WBCoreSerializerFieldMixin, serializers.ReadOnlyField):
    field_type = WBCoreType.TEXT.value


class SlugRelatedField(WBCoreSerializerFieldMixin, serializers.SlugRelatedField):
    field_type = WBCoreType.TEXT.value


class SerializerMethodField(WBCoreSerializerFieldMixin, serializers.SerializerMethodField):
    def __init__(self, method_name=None, field_class=TextField, **kwargs):
        self.field_class = field_class
        self.initkwargs = kwargs
        super().__init__(method_name, **kwargs)

    def get_representation(self, request, field_name):
        field_name, rep = super().get_representation(request, field_name)
        field_name, field_rep = self.field_class(**self.initkwargs).get_representation(request, field_name)

        rep.update(field_rep)
        return field_name, rep



import pytest
from django.apps import apps
from django.db.models.signals import pre_migrate
from wbcore.contrib.example_app.tests.signals import app_pre_migration

pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("geography"))


@pytest.fixture
def chrome_options(chrome_options):
    chrome_options.add_argument("--headless=new")
    chrome_options.add_argument("--disable-extensions")
    return chrome_options



import pytest
from wbcore.enums import Button, Unit


class TestUnit:
    @pytest.mark.parametrize("unit", [Unit.FRACTION, Unit.REM, Unit.PIXEL])
    @pytest.mark.parametrize("value", ["1.0", 1.0, 1])
    def test_unit(self, unit, value):
        assert unit.unit(value) == f"{float(value)}{unit.value}"

    @pytest.mark.parametrize("unit", [Unit.FRACTION, Unit.REM, Unit.PIXEL])
    @pytest.mark.parametrize("value, exception", [("a", ValueError), (None, AssertionError)])
    def test_unit_fail(self, unit, value, exception):
        with pytest.raises(exception):
            unit.unit(value)


class TestButton:
    @pytest.mark.parametrize(
        "method, buttons",
        [
            (
                "buttons",
                [
                    Button.REFRESH.value,
                    Button.NEW.value,
                    Button.DELETE.value,
                    Button.SAVE.value,
                    Button.SAVE_AND_CLOSE.value,
                    Button.SAVE_AND_NEW.value,
                ],
            ),
            (
                "create_buttons",
                [
                    Button.SAVE.value,
                    Button.SAVE_AND_CLOSE.value,
                    Button.SAVE_AND_NEW.value,
                    Button.RESET.value,
                ],
            ),
            (
                "custom_buttons",
                [
                    Button.DROPDOWN.value,
                    Button.HYPERLINK.value,
                    Button.WIDGET.value,
                    Button.ACTION.value,
                ],
            ),
        ],
    )
    def test_methods(self, method, buttons):
        assert getattr(Button, method)() == buttons



import sys
from datetime import datetime
from uuid import UUID

from celery import shared_task
from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.db.models import Q
from django.template import Context, Template, TemplateSyntaxError
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy
from rest_framework.reverse import reverse
from wbcore.contrib.authentication.models import Group, Permission, User
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import base_domain, create_notification_type
from wbcore.contrib.workflow.models import Process, ProcessStep
from wbcore.contrib.workflow.models.transition import Transition
from wbcore.contrib.workflow.sites import workflow_site
from wbcore.contrib.workflow.utils import get_model_serializer_class_for_instance
from wbcore.models import WBModel
from wbcore.utils.html import convert_html2text
from wbcore.utils.string_loader import StringSourceLoader


class Step(WBModel):
    """One step in the workflow"""

    class StepType(models.TextChoices):
        STARTSTEP = "StartStep", _("Start Step")
        USERSTEP = "UserStep", _("User Step")
        DECISIONSTEP = "DecisionStep", _("Decision Step")
        SPLITSTEP = "SplitStep", _("Split Step")
        JOINSTEP = "JoinStep", _("Join Step")
        EMAILSTEP = "EmailStep", _("Email Step")
        SCRIPTSTEP = "ScriptStep", _("Script Step")
        FINISHSTEP = "FinishStep", _("Finish Step")

    name = models.CharField(max_length=128, verbose_name=_("Name"))
    workflow = models.ForeignKey(
        to="workflow.Workflow", verbose_name=_("Workflow"), on_delete=models.CASCADE, related_name="associated_steps"
    )
    code = models.PositiveIntegerField(verbose_name=_("Code"))
    status = models.CharField(
        max_length=64,
        verbose_name=_("Status"),
        help_text=_(
            "The status that will be set in the attached instance's status field upon transitioning to this step. Only applicable if attached model is set."
        ),
        null=True,
        blank=True,
    )
    step_type = models.CharField(max_length=64, default="", choices=StepType.choices, verbose_name=_("Step Type"))
    permission = models.ForeignKey(
        to=Permission,
        verbose_name=_("Permission"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="related_steps",
        help_text=_("Define which permission is needed to be able to view this step being executed."),
    )

    def get_outgoing_transitions(self) -> models.QuerySet[Transition]:
        return self.outgoing_transitions.all()

    def get_incoming_transitions(self) -> models.QuerySet[Transition]:
        return self.incoming_transitions.all()

    def get_previous_steps(self) -> list["Step"]:
        previous_steps = []
        for transition in self.get_incoming_transitions():
            previous_steps.append(transition.from_step)
        return previous_steps

    def get_all_valid_outgoing_transitions(self, process_step: ProcessStep) -> list[Transition]:
        transition_list = []
        for transition in self.get_outgoing_transitions():
            if transition.all_conditions_satisfied(process_step):
                transition_list.append(transition)
        return transition_list

    def get_assigned_group(self) -> None:
        return None

    def get_assigned_user(self) -> None:
        return None

    def get_casted_step(self) -> "Step":
        """Casts the step into its child representative"""

        return getattr(sys.modules[__name__], self.step_type).objects.get(pk=self.pk)

    def user_can_see_step(self, user: User) -> bool:
        """Returns true if the user or any of the user groups he is part of has the permission to see the process step being executed and choose the next step

        Args:
            user (User): The user instance

        Returns:
            bool: True if the user has the permission and False otherwise
        """

        return (
            (not self.permission)
            or user.is_superuser
            or self.permission in user.user_permissions.all()
            or user.groups.filter(permissions=self.permission).exists()
        )

    def set_finished(self, process_step: ProcessStep):
        process_step.state = ProcessStep.StepState.FINISHED
        process_step.save()
        self.finish(process_step)

    def set_failed(self, process_step: ProcessStep, error_message: str):
        process_step.state = ProcessStep.StepState.FAILED
        # We don't want to overwrite a previous error message
        if not process_step.error_message:
            process_step.error_message = error_message
        process_step.save()

        # Check if the process can still finish
        process_can_finish.delay(process_step.process.pk)

    def set_canceled(self, process_step: ProcessStep):
        process_step.state = ProcessStep.StepState.CANCELED
        process_step.save()
        self.finish(process_step)

    def execute_single_next_step(self, process_step: ProcessStep):
        """Finds the next step and executes it. Fails if more than one step or no step can be transitioned to.

        Args:
            process_step (ProcessStep): Current process step
        """

        if len(valid_transitions := self.get_all_valid_outgoing_transitions(process_step)) == 1:
            Step.start_next_step(process_step, valid_transitions[0])
        elif not valid_transitions:
            self.set_failed(process_step, gettext("No valid outgoing transitions found for this step!"))
        else:
            self.set_failed(process_step, gettext("More than one possible outgoing transition found for this step!"))

    def run(self, process_step: ProcessStep):
        self.set_finished(process_step)

    def finish(self, process_step: ProcessStep):
        if process_step.state in [ProcessStep.StepState.FINISHED, ProcessStep.StepState.CANCELED]:
            # Set finished field
            process_step.finished = datetime.now()
            process_step.save()

    def __str__(self) -> str:
        return self.name

    def save(self, *args, **kwargs):
        self.step_type = self.__class__.__name__
        super().save(*args, **kwargs)

    @classmethod
    def start_next_step(cls, process_step: ProcessStep, transition: Transition):
        """Starts the next step of the transition and finishes the previous one

        Args:
            process_step (ProcessStep): The current process step
            transition (Transition): The transition that should be executed
        """

        next_step: Step = transition.to_step.get_casted_step()
        previous_step: Step = transition.from_step.get_casted_step()
        previous_step.set_finished(process_step)
        activate_step.delay(next_step.pk, process_step.process.pk)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:step"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:steprepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    class Meta:
        verbose_name = _("Step")
        verbose_name_plural = _("Steps")
        constraints = [
            models.UniqueConstraint(fields=["name", "workflow"], name="unique_name_workflow"),
            models.UniqueConstraint(fields=["code", "workflow"], name="unique_code_workflow"),
        ]


class StartStep(Step):
    """The first step in a workflow"""

    def run(self, process_step: ProcessStep):
        self.execute_single_next_step(process_step)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:startstep"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:startsteprepresentation-list"

    class Meta:
        verbose_name = _("Start Step")
        verbose_name_plural = _("Start Steps")


class UserStep(Step):
    """A workflow step that needs an assigned user or user group to pick the next transition."""

    assignee = models.ForeignKey(
        to=User,
        verbose_name=_("Assignee"),
        on_delete=models.SET_NULL,
        related_name="assigned_workflow_steps",
        null=True,
        blank=True,
    )
    notify_user = models.BooleanField(verbose_name=_("Notify User"), default=True)
    group = models.ForeignKey(
        to=Group,
        verbose_name=_("Group"),
        on_delete=models.SET_NULL,
        related_name="assigned_workflow_steps",
        null=True,
        blank=True,
    )
    assignee_method = models.CharField(
        max_length=64,
        verbose_name=_("Assignee Method"),
        blank=True,
        null=True,
    )
    display = models.ForeignKey(
        to="workflow.Display",
        verbose_name=_("Display"),
        on_delete=models.SET_NULL,
        related_name="user_steps",
        null=True,
        blank=True,
    )
    kwargs = models.JSONField(verbose_name=_("Kwargs"), null=True, blank=True)

    def get_assigned_group(self) -> None | Group:
        return self.group

    def get_assigned_user(self) -> None | User:
        return self.assignee

    def execute_assignee_method(self, process_step: ProcessStep, method_name: str, **kwargs):
        for choice in workflow_site.assignees_methods:
            if method_name == choice[0]:
                process_step.assignee = choice[1](process_step, **kwargs)
                process_step.save()
                break

    def run(self, process_step: ProcessStep):
        if method := self.assignee_method:
            if self.kwargs:
                self.execute_assignee_method(process_step, method, **self.kwargs)
            else:
                self.execute_assignee_method(process_step, method)

        if not process_step.state == ProcessStep.StepState.FAILED:
            if self.get_all_valid_outgoing_transitions(process_step):
                if all_assignees := process_step.get_all_assignees():
                    if self.notify_user:
                        self.notify_assignees(all_assignees, process_step)
                else:
                    self.set_failed(process_step, gettext("No assignees selected!"))
            else:
                self.set_failed(process_step, gettext("No valid outgoing transitions found for this step!"))

    def set_failed(self, process_step: ProcessStep, error_message: str):
        process_step.state = ProcessStep.StepState.FAILED
        # We don't want to overwrite a previous error message
        if not process_step.error_message:
            process_step.error_message = error_message
        process_step.save()
        if self.notify_user and (all_assignees := process_step.get_all_assignees()):
            self.notify_assignees(all_assignees, process_step, error_message)

        # Check if the process can still finish
        process_can_finish.delay(process_step.process.pk)

    def notify_assignees(self, assignees: list[User], process_step: ProcessStep, error_message: str = ""):
        if error_message:
            code = "workflow.userstep.notify_failed_step"
            title = gettext("Assigned Workflow Step Failed")
            body = gettext(
                "A workflow step you were assigned to just failed with the error message '{}'. Please take appropriate action."
            ).format(error_message)
        else:
            code = "workflow.userstep.notify_next_step"
            title = gettext("Workflow Step Awaiting Your Decision")
            body = gettext("You were assigned to a workflow step. Please select the next step.")

        for user in assignees:
            if user.is_active and self.user_can_see_step(user):
                send_notification(
                    code=code,
                    title=title,
                    body=body,
                    user=user,
                    endpoint=reverse(f"{process_step.get_endpoint_basename()}-detail", args=[process_step.id]),
                )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:userstep"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:usersteprepresentation-list"

    class Meta:
        verbose_name = _("User Step")
        verbose_name_plural = _("User Steps")
        constraints = [
            models.CheckConstraint(
                check=~Q(assignee__isnull=False, group__isnull=False),
                name="check_not_both_assignee_group",
            ),
        ]
        notification_types = [
            create_notification_type(
                code="workflow.userstep.notify_next_step",
                title=gettext("Workflow Step Assignment Notification"),
                help_text=gettext(
                    "Notification for all assigned users that can transition a workflow step to the next step"
                ),
            ),
            create_notification_type(
                code="workflow.userstep.notify_failed_step",
                title=gettext("Failed Workflow Step Notification"),
                help_text=gettext("Notification for all assigned users of a failed workflow step"),
            ),
        ]


class DecisionStep(Step):
    """A background step that checks all conditions and activates the first transition that satisfies all of its conditions.
    Only really makes sense in conjunction with conditions that disable every transition but one.
    """

    def get_first_valid_transition(self, process_step: ProcessStep) -> Transition | None:
        for transition in self.get_outgoing_transitions():
            if transition.all_conditions_satisfied(process_step):
                return transition
        return None

    def run(self, process_step: ProcessStep):
        if transition := self.get_first_valid_transition(process_step):
            Step.start_next_step(process_step, transition)
        else:
            self.set_failed(process_step, gettext("No valid outgoing transition found for this step!"))

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:decisionstep"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:decisionsteprepresentation-list"

    class Meta:
        verbose_name = _("Decision Step")
        verbose_name_plural = _("Decision Steps")


class SplitStep(Step):
    """A background step that activates all transitions that meet all of their conditions"""

    def run(self, process_step: ProcessStep):
        if valid_transitions := self.get_all_valid_outgoing_transitions(process_step):
            self.set_finished(process_step)
            for transition in valid_transitions:
                next_step: Step = transition.to_step
                activate_step.delay(next_step.pk, process_step.process.pk)
        else:
            self.set_failed(process_step, gettext("No valid outgoing transitions found for this step!"))

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:splitstep"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:splitsteprepresentation-list"

    class Meta:
        verbose_name = _("Split Step")
        verbose_name_plural = _("Split Steps")


class JoinStep(Step):
    """A background step that joins multiple steps into one. Waits until all incoming transitions are done by default.
    If wait_for_all is False the first incoming transition cancels all other active incoming transitions."""

    wait_for_all = models.BooleanField(
        default=True,
        verbose_name=_("Wait For All"),
        help_text=_("If False, will cancel all other incoming process steps as soon the first reaches this step."),
    )

    def cancel_if_leading_to_self(self, step: Step, process_step: ProcessStep):
        """Recursively goes through all outgoing transitions of the provided step. Cancels the provided process step as soon as a transition that leads to the current split step is found.

        Args:
            step (Step): The step from which all outgoing transitions need to be investigated
            process_step (ProcessStep): The related process step that would need to be canceled
        """

        for transition in step.outgoing_transitions.all():
            if transition.to_step.pk == self.pk:
                process_step.step.set_canceled(process_step)
            else:
                self.cancel_if_leading_to_self(transition.to_step, process_step)

    def run(self, process_step: ProcessStep):
        process_step.state = ProcessStep.StepState.WAITING
        process_step.save()

        if self.wait_for_all:
            # Look for incoming transitions with no process step (not activated yet) or not finished
            if not Transition.objects.filter(
                Q(to_step=self)
                & Q(
                    Q(from_step__process_steps__isnull=True)
                    | Q(
                        ~Q(
                            from_step__process_steps__state__in=[
                                ProcessStep.StepState.FINISHED,
                                ProcessStep.StepState.CANCELED,
                            ]
                        )
                        & Q(from_step__process_steps__process=process_step.process)
                    )
                )
            ).exists():
                self.execute_single_next_step(process_step)
        else:
            # Set every unfinished process step that at some point leads to this step to canceled
            for unfinished_process_step in ProcessStep.objects.filter(process=process_step.process).exclude(
                Q(
                    Q(pk=process_step.pk)
                    | Q(
                        state__in=[
                            ProcessStep.StepState.FINISHED,
                            ProcessStep.StepState.CANCELED,
                            ProcessStep.StepState.FAILED,
                        ]
                    )
                )
            ):
                self.cancel_if_leading_to_self(unfinished_process_step.step, unfinished_process_step)

            self.execute_single_next_step(process_step)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:joinstep"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:joinsteprepresentation-list"

    class Meta:
        verbose_name = _("Join Step")
        verbose_name_plural = _("Join Steps")


class ScriptStep(Step):
    """A background step that will execute a provided python script."""

    script = models.TextField(verbose_name=_("Script"))

    def run(self, process_step: ProcessStep):
        try:
            loader = StringSourceLoader(self.script)
            loader.load_module().run(process_step)
        except Exception as e:
            self.set_failed(process_step, gettext("Executing script failed: {}".format(e)))
            return
        loader.cleanup()
        self.execute_single_next_step(process_step)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:scriptstep"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:scriptsteprepresentation-list"

    class Meta:
        verbose_name = _("Script Step")
        verbose_name_plural = _("Script Steps")


class EmailStep(Step):
    """A background step that sends an email to a list of email contacts."""

    template = models.FileField(
        max_length=256,
        verbose_name=_("Template"),
        upload_to="workflow/templates/Test_Templates.txt",
    )
    subject = models.CharField(
        max_length=128,
        verbose_name=_("Subject"),
    )
    to = models.ManyToManyField(
        "directory.EmailContact",
        related_name="workflow_step_receiver",
        verbose_name=pgettext_lazy("Email context", "To"),
    )
    cc = models.ManyToManyField(
        "directory.EmailContact", related_name="workflow_steps_cc", verbose_name=_("CC"), blank=True
    )
    bcc = models.ManyToManyField(
        "directory.EmailContact", related_name="workflow_steps_bcc", verbose_name=_("BCC"), blank=True
    )

    def run(self, process_step: ProcessStep):
        context = {
            "process_step_endpoint": base_domain()
            + reverse(f"{process_step.get_endpoint_basename()}-detail", args=[process_step.id]),
            "process_step": process_step,
        }
        try:
            rendered_html = Template(self.template.file).render(Context(context))
            msg = EmailMultiAlternatives(
                self.subject,
                convert_html2text(rendered_html),
                settings.DEFAULT_FROM_EMAIL,
                to=list(self.to.values_list("address", flat=True)),
                cc=list(self.cc.values_list("address", flat=True)),
                bcc=list(self.bcc.values_list("address", flat=True)),
            )
            msg.attach_alternative(rendered_html, "text/html")
            msg.send()
            self.execute_single_next_step(process_step)
        except TemplateSyntaxError:
            self.set_failed(process_step, gettext("Error in template syntax!"))

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:emailstep"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:emailsteprepresentation-list"

    class Meta:
        verbose_name = _("Email Step")
        verbose_name_plural = _("Email Steps")


class FinishStep(Step):
    """The last step in a workflow"""

    write_preserved_instance = models.BooleanField(
        default=False,
        verbose_name=_("Write Preserved Instance"),
        help_text=_(
            "Writes the instance data backup onto the instance restoring its state to the beginning of the workflow."
        ),
    )

    def run(self, process_step: ProcessStep):
        # Finish if there are no running process steps for this process
        if (
            ProcessStep.objects.filter(
                process=process_step.process, state__in=[ProcessStep.StepState.WAITING, ProcessStep.StepState.ACTIVE]
            )
            .exclude(id=process_step.pk)
            .exists()
        ):
            self.set_failed(process_step, gettext("There are process steps still running for this workflow!"))
        else:
            self.set_finished(process_step)

    def finish(self, process_step: ProcessStep):
        super().finish(process_step)

        if process_step.state == ProcessStep.StepState.FINISHED:
            # Save preserved instance
            if self.write_preserved_instance and (preserved_data := process_step.process.preserved_instance):
                attached_instance = process_step.process.instance
                serializer = get_model_serializer_class_for_instance(attached_instance)(data=preserved_data)
                serializer.is_valid()
                validated_data: dict = serializer.validated_data
                serializer.update(attached_instance, validated_data)

            # Finish entire process if process step is finished
            process: Process = process_step.process
            process.finished = datetime.now()
            process.state = Process.ProcessState.FINISHED
            process.save()

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:finishstep"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:finishsteprepresentation-list"

    class Meta:
        verbose_name = _("Finish Step")
        verbose_name_plural = _("Finish Steps")


@shared_task()
def activate_step(step_id: int, process_id: UUID):
    step = Step.objects.get(pk=step_id).get_casted_step()
    process = Process.objects.get(pk=process_id)

    # We need to prevent creating a duplicate join step
    if (
        step.step_type == Step.StepType.JOINSTEP
        and (
            process_step_qs := ProcessStep.objects.filter(
                process=process,
                step=step,
                permission=step.permission,
                status=step.status,
                state=ProcessStep.StepState.WAITING,
                finished__isnull=True,
                group=step.get_assigned_group(),
                assignee=step.get_assigned_user(),
            )
        ).count()
        == 1
    ):
        process_step: ProcessStep = process_step_qs.first()
    else:
        process_step = ProcessStep.objects.create(
            process=process,
            step=step,
            permission=step.permission,
            status=step.status,
            group=step.get_assigned_group(),
            assignee=step.get_assigned_user(),
        )

    # Set status in attached model
    if step.status and (attached_instance := process_step.process.instance):
        field_name = process_step.process.workflow.status_field
        if hasattr(attached_instance, field_name):
            if attached_instance.__class__._meta.get_field(field_name).get_internal_type() == "CharField":
                setattr(attached_instance, field_name, step.status)
                attached_instance.save()

    step.run(process_step)


def _check_previous_steps_failed(step: Step, process: Process, check_all: bool = False):
    """Recursively checks if the previous steps of a step have failed and fails the process if they were critical.

    Args:
        step (Step): The step to be checked
        process (Process): The process to be checked
        check_all (bool, optional): Indicates if all of a step's previous steps need to fail in order to fail the entire process. Defaults to False.
    """

    previous_steps = step.get_previous_steps()
    failed_steps = 0
    for previous_step in previous_steps:
        if (process_steps_for_process_qs := previous_step.process_steps.filter(process=process)).exists():
            if process_steps_for_process_qs.first().state == ProcessStep.StepState.FAILED:
                if check_all:
                    # Only fail process if ALL previous steps have failed
                    failed_steps += 1
                    if failed_steps == len(previous_steps):
                        process.set_failed()
                        return
                else:
                    process.set_failed()
                    return
        else:
            for step in previous_step.get_previous_steps():
                if previous_step.step_type == Step.StepType.JOINSTEP:
                    if previous_step.get_casted_step().wait_for_all is False:
                        _check_previous_steps_failed(step, process, check_all=True)
                    # Join steps with wait_for_all=True will be checked by main for loop
                    continue
                else:
                    _check_previous_steps_failed(step, process)


@shared_task()
def process_can_finish(process_id: UUID):
    """Checks if all critical steps for a process can still be reached. Fails the process if this is not the case.

    Args:
        process_id (UUID): ID of the process to be checked
    """

    process = Process.objects.get(pk=process_id)

    # We can immediately fail the entire process when there are no more active process steps
    if not process.process_steps.filter(
        state__in=[ProcessStep.StepState.WAITING, ProcessStep.StepState.ACTIVE]
    ).exists():
        process.set_failed()

    # Otherwise we need to check if all unfinished join steps that need several inputs can still be reached
    else:
        waiting_join_steps = JoinStep.objects.filter(
            Q(workflow=process.workflow)
            & Q(wait_for_all=True)
            & Q(
                Q(process_steps__isnull=True)
                | Q(
                    Q(process_steps__state__in=[ProcessStep.StepState.ACTIVE, ProcessStep.StepState.WAITING])
                    & Q(process_steps__process=process)
                )
            )
        )
        for join_step in waiting_join_steps:
            _check_previous_steps_failed(join_step, process)



from rest_framework.reverse import reverse
from wbcore.contrib.documents.models import ShareableLinkAccess
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ShareableLinkEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        base_url = "wbcore:documents:link-list"

        if doc_id := self.request.GET.get("document", None):
            return f"{reverse(base_url, request=self.request)}?document={doc_id}"

        return reverse(base_url, request=self.request)

    def get_create_endpoint(self, **kwargs):
        base_url = "wbcore:documents:link-list"

        if doc_id := self.request.GET.get("document", None):
            return f"{reverse(base_url, request=self.request)}?document={doc_id}"

        return reverse(base_url, request=self.request)

    def get_delete_endpoint(self, **kwargs):
        try:
            if ShareableLinkAccess.objects.filter(shareable_link=self.view.get_object()).exists():
                return None
        except AssertionError:
            pass
        return self.get_endpoint()


class ShareableLinkAccessEndpointConfig(EndpointViewConfig):
    def get_list_endpoint(self, **kwargs):
        base_url = "wbcore:documents:linkaccess-list"

        if link_id := self.request.GET.get("shareable_link", None):
            return f"{reverse(base_url, request=self.request)}?shareable_link={link_id}"

        return reverse(base_url, request=self.request)

    def get_create_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        return None



from datetime import date, timedelta
from decimal import Decimal

import pytest
from django.contrib.postgres.aggregates import ArrayAgg
from django.db.models import Q
from django.forms.models import model_to_dict
from rest_framework import status
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.currency.factories import CurrencyFactory, CurrencyFXRatesFactory
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.currency.viewsets import (
    CurrencyModelViewSet,
    CurrencyRepresentationViewSet,
)


@pytest.mark.viewset_tests
@pytest.mark.currency_representation_viewset_tests
@pytest.mark.with_db
@pytest.mark.django_db
class TestCurrencyRepresentationViewSets:
    @pytest.fixture()
    def test_request(self):
        factory = APIRequestFactory()
        request = factory.get("")
        request.user = SuperUserFactory()
        return request

    def test_currency_list_representation_viewset(self, test_request, currency_factory):
        # Arrange
        currency_factory.create_batch(2)
        vs = CurrencyRepresentationViewSet.as_view({"get": "list"})
        # Act
        response = vs(test_request)
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data
        assert len(response.data["results"]) == 2

    def test_currency_representation_viewset(self, test_request, currency_factory):
        # Arrange
        euro = currency_factory(symbol="€", key="EUR", title="Euro")
        vs = CurrencyRepresentationViewSet.as_view({"get": "retrieve"})
        # Act
        response = vs(test_request, pk=euro.pk)
        instance = response.data.get("instance")
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert instance.get("id") == euro.id
        assert instance.get("key") == euro.key
        assert instance.get("title") == euro.title


@pytest.mark.viewset_tests
@pytest.mark.currency_viewset_tests
@pytest.mark.with_db
@pytest.mark.django_db
class TestCurrencyViewSets:
    @pytest.fixture()
    def api_rf(self):
        factory = APIRequestFactory()
        return factory

    @pytest.fixture()
    def su(self):
        super_user = SuperUserFactory()
        return super_user

    @pytest.fixture()
    def get_request(self, api_rf, su):
        request = api_rf.get("")
        request.user = su
        return request

    @pytest.fixture()
    def currency(self):
        currency = CurrencyFactory()
        return currency

    @pytest.fixture()
    def currencies(self):
        currency_a = CurrencyFactory()
        currency_b = CurrencyFactory()
        currency_c = CurrencyFactory()
        return [currency_a, currency_b, currency_c]

    def test_get_instance_request(self, get_request, currency):
        # Arrange
        vs = CurrencyModelViewSet.as_view({"get": "retrieve"})
        # Act
        response = vs(get_request, pk=currency.pk)
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert not response.data.get("results")
        assert response.data["instance"]["id"] == currency.id

    def test_get_list_request(self, get_request):
        # Arrange
        CurrencyFactory.create_batch(3)
        vs = CurrencyModelViewSet.as_view({"get": "list"})
        # Act
        response = vs(get_request)
        # Assert
        results = response.data.get("results")
        assert response.status_code == status.HTTP_200_OK
        assert not response.data.get("instance")
        assert len(results) == 3

    def test_delete_instance(self, api_rf, su, currency):
        # Arrange
        request = api_rf.delete("")
        request.user = su
        view = CurrencyModelViewSet.as_view({"delete": "destroy"})
        # Act
        response = view(request, pk=currency.id)
        # Assert
        assert response.status_code == status.HTTP_204_NO_CONTENT
        assert Currency.objects.count() == 0

    def test_update_instance(self, api_rf, su, currency):
        # Arrange
        new_currency = model_to_dict(CurrencyFactory.build())
        # Add missing values to the dict, because we cannot update data with 'None' values
        new_currency["id"] = currency.id
        new_currency["import_source"] = ""
        request = api_rf.put("", data=new_currency)
        request.user = su
        view = CurrencyModelViewSet.as_view({"put": "update"})
        # Act
        response = view(request, pk=currency.id)
        instance = response.data.get("instance")
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert instance["id"] == currency.id
        assert instance["key"] == new_currency["key"]
        assert instance["key"] != currency.key

    @pytest.mark.parametrize("title", ["New_title"])
    def test_patch_instance(self, api_rf, su, currency, title):
        # Arrange
        request = api_rf.patch("", data={"title": title})
        request.user = su
        # Act
        view = CurrencyModelViewSet.as_view({"patch": "partial_update"})
        response = view(request, pk=currency.id)
        # Assert
        assert response.status_code == status.HTTP_200_OK
        assert response.data["instance"]["title"] == title
        assert currency.title != title
        currency.refresh_from_db()  # The patch changes are only visible after a refresh
        assert currency.title == title

    def test_get_queryset_no_fx_rates(self, currencies):
        viewset = CurrencyModelViewSet()
        qs = viewset.get_queryset()
        last_rate_date = date.today() - timedelta(days=30)
        expected_queryset = Currency.objects.annotate(
            fx_rates_date=ArrayAgg("fx_rates__date", filter=Q(fx_rates__date__gt=last_rate_date)),
            fx_rates_value=ArrayAgg("fx_rates__value", filter=Q(fx_rates__date__gt=last_rate_date)),
        )
        assert list(qs) == list(expected_queryset)
        for currency in qs:
            assert currency.fx_rates_date is None
            assert currency.fx_rates_value is None

    def test_get_queryset_with_fx_rates(self, currencies):
        recent_fx_rate_date, mid_fx_rate_date, past_cutoff_fx_rate_date = (
            date.today() - timedelta(days=5),
            date.today() - timedelta(days=20),
            date.today() - timedelta(days=35),
        )
        value_a, value_b, value_c = (
            Decimal("1.10"),
            Decimal("0.85"),
            Decimal("1.20"),
        )

        fx_rate_data = [
            {"date": recent_fx_rate_date, "value": value_a},
            {"date": mid_fx_rate_date, "value": value_b},
            {"date": past_cutoff_fx_rate_date, "value": value_c},
        ]

        for currency in currencies:
            for data in fx_rate_data:
                CurrencyFXRatesFactory(currency=currency, date=data["date"], value=data["value"])

        viewset = CurrencyModelViewSet()
        qs = viewset.get_queryset()
        expected_rates = [value_a, value_b]
        expected_dates = [recent_fx_rate_date, mid_fx_rate_date]
        assert len(qs) == 3
        for currency in qs:
            assert currency.fx_rates_date == expected_dates
            assert currency.fx_rates_value == expected_rates



from contextlib import suppress

from django.core.exceptions import FieldDoesNotExist
from rest_framework import serializers
from rest_framework.fields import empty
from rest_framework.relations import ManyRelatedField, PKOnlyObject
from rest_framework.request import Request

from .list import ListFieldMixin
from .mixins import WBCoreSerializerFieldMixin
from .types import WBCoreType


class WBCoreManyRelatedField(ListFieldMixin, WBCoreSerializerFieldMixin, ManyRelatedField):
    def __init__(self, *args, **kwargs):
        required = kwargs.get("required", True)
        if not required:
            kwargs["allow_null"] = True
        super().__init__(*args, **kwargs)

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        # The `many=True` parameter on the related field returns this class,
        # where `children_relation` is directly bound to it. However, the context
        # processor does not yet have access to the request, so the `view` property
        # is currently `None`. To handle this, we need to explicitly re-evaluate
        # the `read_only` property for the `children_relation` when the binding
        # occurs from the List serializer to this instance.
        self.child_relation.context["view"] = self.view
        self.child_relation._evaluate_read_only(field_name, parent)
        if not self.child_relation.read_only and hasattr(self.child_relation, "_queryset"):
            setattr(self.child_relation, "queryset", self.child_relation._queryset)

    def get_representation(self, request: Request, field_name: str) -> tuple[str, dict]:
        key, representation = self.child_relation.get_representation(request, field_name)
        representation["multiple"] = True
        if not representation.get("label", None) and self.label:
            representation["label"] = self.label
        return key, representation


class PrimaryKeyRelatedField(WBCoreSerializerFieldMixin, serializers.PrimaryKeyRelatedField):
    MANY_RELATION_KWARGS = (
        "read_only",
        "write_only",
        "required",
        "default",
        "initial",
        "source",
        "label",
        "help_text",
        "style",
        "error_messages",
        "allow_empty",
        "html_cutoff",
        "html_cutoff_text",
        "allow_null",
    )

    def __init__(self, *args, queryset=None, read_only=False, **kwargs):
        self.field_type = kwargs.pop("field_type", WBCoreType.SELECT.value)
        if callable(read_only) and queryset is not None:
            self._queryset = queryset  # we unset any given queryset to be compliant with the RelatedField assertion
            queryset = None
        super().__init__(*args, queryset=queryset, read_only=read_only, **kwargs)

    def __new__(cls, *args, **kwargs):
        kwargs["style"] = {"base_template": "input.html"}
        return super().__new__(cls, *args, **kwargs)

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        # In case we had to unset the queryset attribute because read_only was a callable, we reinstate it here.
        if not self.read_only and hasattr(self, "_queryset"):
            setattr(self, "queryset", self._queryset)

    @classmethod
    def many_init(cls, *args, **kwargs):
        list_kwargs = {"child_relation": cls(*args, **kwargs)}
        for key in kwargs:
            if key in cls.MANY_RELATION_KWARGS:
                list_kwargs[key] = kwargs[key]
        return WBCoreManyRelatedField(**list_kwargs)

    def run_validation(self, data=empty):
        if (data is empty) and (view := self.parent.context.get("view", None)):
            parent_model_id = view.kwargs.get(f"{self.field_name}_id")
            if parent_model_id:
                data = parent_model_id

        return super().run_validation(data)

    def get_representation(self, request: Request, field_name: str) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        if default := representation.get("default", None):
            if isinstance(default, list):
                representation["default"] = [getattr(d, "id", d) for d in default]
            elif id := getattr(default, "id", None):
                representation["default"] = id
        return key, representation

    def to_representation(self, value):
        # In case we annotate the representation, we need to ensure that the value is an object
        if isinstance(value, (list, tuple, set)):
            return [self.to_representation(d) for d in value]
        try:
            if isinstance(value, str):
                try:
                    value = int(value)
                except ValueError:
                    value = None
            if isinstance(value, int):
                value = PKOnlyObject(value)
            return super().to_representation(value)
        except Exception:
            pass
        return None

    def get_queryset(self):
        """
        Allow queryset to be callable
        """
        if callable(self.queryset):
            self.queryset = self.queryset()
        qs = super().get_queryset()

        # if limite_choices_to is defined on the model field, we use it to restrict the related queryset
        with suppress(AttributeError, FieldDoesNotExist):
            if limit_choices_to := getattr(
                self.parent.Meta.model._meta.get_field(self.field_name), "_limit_choices_to", None
            ):
                if isinstance(limit_choices_to, dict):
                    qs = qs.filter(**limit_choices_to)
                else:
                    qs = qs.filter(limit_choices_to)
        return qs


class ListSerializer(WBCoreSerializerFieldMixin, serializers.ListSerializer):
    """
    A Wrapper around the normal DRF ListSerializer which also return the child representation
    """

    def get_representation(self, request: Request, field_name: str) -> tuple[str, dict]:
        _, representation = self.child.get_representation(request, field_name)
        representation["multiple"] = True
        related_key = self.get_related_key()
        representation["related_key"] = related_key

        # TODO: improve and don't monkeypatch
        rv = related_key if "." not in related_key else related_key.replace(".", "__")
        return rv, representation



from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

RELATIONSHIPTYPE_MENUITEM = MenuItem(
    label=_("Relationship Types"),
    endpoint="wbcore:directory:relationship-type-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["directory.view_relationshiptype"],
    ),
    add=MenuItem(
        label=_("Create Relationship Type"),
        endpoint="wbcore:directory:relationship-type-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user),
            permissions=["directory.add_relationshiptype"],
        ),
    ),
)

CLIENTMANAGER_MENUITEM = MenuItem(
    label=_("Client Manager Relationships"),
    endpoint="wbcore:directory:clientmanagerrelationship-list",
    permission=ItemPermission(permissions=["directory.view_clientmanagerrelationship"]),
    add=MenuItem(
        label=_("Create Client Manager Relationship"),
        endpoint="wbcore:directory:clientmanagerrelationship-list",
        permission=ItemPermission(permissions=["directory.add_clientmanagerrelationship"]),
    ),
)



import pathlib

import factory
import yaml

from .models import Currency, CurrencyFXRates

with open(pathlib.Path(__file__).parent.joinpath("fixtures").joinpath("currency.yaml"), "r") as yaml_file:
    currency_dict = yaml.load(yaml_file, Loader=yaml.CLoader)


class CurrencyFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Currency
        django_get_or_create = ["key"]

    title = factory.Iterator(currency_dict, getter=lambda c: c["fields"]["title"])
    symbol = factory.Iterator(currency_dict, getter=lambda c: c["fields"]["symbol"])
    key = factory.Iterator(currency_dict, getter=lambda c: c["fields"]["key"])

    @classmethod
    def full_db(cls):
        for _ in range(len(currency_dict)):
            cls()


class CurrencyUSDFactory(CurrencyFactory):
    title = "US Dollar"
    symbol = "$"
    key = "USD"


class CurrencyFXRatesFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = CurrencyFXRates

    date = factory.Faker("date_object")
    currency = factory.SubFactory(CurrencyFactory)
    value = factory.Faker("pydecimal", min_value=1, max_value=999999, right_digits=6)



from wbcore import serializers
from wbcore.contrib.example_app.models import Role


class RoleModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = Role
        fields = "__all__"


class RoleRepresentationSerializer(serializers.RepresentationSerializer):
    _detail = serializers.HyperlinkField(reverse_name="example_app:role-detail")

    class Meta:
        model = Role
        fields = ("id", "title", "_detail")



from django.contrib import admin
from wbcore.contrib.workflow.models import Data, DataValue


@admin.register(Data)
class DataAdmin(admin.ModelAdmin):
    search_fields = ("label",)
    list_display = ("id", "label", "data_type", "workflow", "help_text", "required", "default")


@admin.register(DataValue)
class DataValueAdmin(admin.ModelAdmin):
    search_fields = ("value",)
    list_display = ("id", "value", "data")
